!
! Module: polygon
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains several routines to do with coordinate transformations.
! 
! Comments: TODO:
! 
! $Id: coordinate_mod.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module polygon

  use coordinate ! for pointType
  use ssc_misc   ! for isOdd (maybe for other things??)
!!!  use math_utils

  implicit none 

  private 

  public areaOfPoly
  public regrid
!  public pointIsInsidePoly

  type :: localPolyType
    type(pointType) :: pos
  end type localPolyType
  type :: startPointType
    type(pointType) :: pos, squareLo, squareHi
    integer :: squareXi, squareYi
    logical :: followPoly
  end type startPointType

!logical :: masterTestFlag = .true.

contains 

  !---------------------------------------------------------------------
  subroutine regrid(oldPixelCorners, oldImage, newImage&
  , status)
  !---------------------------------------------------------------------
  ! This sub is intended to allow rebinning of an image from one coordinate
  ! system to another. Now, we define here an image as a two-dimensional
  ! array of uniform rectangular pixels. If we change the coordinate system,
  ! this image becomes distorted. It is nice to be able to convert it once
  ! again to an array of rectangular pixels, but this time in the new
  ! coordinate system. This involves taking each of the old, distorted
  ! pixels and dividing its contents up among the new pixels. It is assumed
  ! here that (i) the distorted pixel still has straight 'sides', ie it is
  ! a quadrilateral (a polygon for generality); (ii) that the value within
  ! the old pixel is evenly distributed through the pixel (this assumption
  ! MAY BE INVALID for extremely nonlinear distortions); (iii) that the
  ! distorted pixel is not folded over - ie that no two of its sides cross.
  !
  ! The relation between the old and new coordinate systems is here entirely contained within the input array oldPixelCorners. This gives the coordinates of each pixel in the array oldImage, expressed in the wcs 'pixel' system of newImage. What does 'wcs pixel system' mean? It means that the width and height of pixels in newImage are both equal to 1.0 and the centre of the pixel newImage(1,1) is at (1.0, 1.0).
  !---------------------------------------------------------------------

    implicit none

    real,           intent(in)            :: oldImage(:,:)
    type(pointType),intent(in)            :: oldPixelCorners(size(oldImage,1)+1&
                                                            ,size(oldImage,2)+1)
    real,           intent(out)           :: newImage(:,:)
    integer,        intent(out), optional :: status

    type(localPolyType) :: poly(4)
    integer :: parity
    integer :: oldXi, oldXsize, oldYi, oldYsize,&
      newXsize, newYsize, foreignStatus
    type(pointType) :: minCornerNew, maxCornerNew
!, deltaNew
    real :: totalPolyArea

    !-------------------------------------------------------------------
    ! Start

    status = 0

    oldXsize = size(oldImage, 1)
    oldYsize = size(oldImage, 2)
    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)

    if (newXsize .le. 0 .or. newYsize .le. 0) then
      status = ibset(status, 2)
  return
    end if

    minCornerNew%x = 0.5
    minCornerNew%y = 0.5
    maxCornerNew%x = newXsize + 0.5
    maxCornerNew%y = newYsize + 0.5

!    deltaNew%x = maxCornerNew%x - minCornerNew%x
!    deltaNew%y = maxCornerNew%y - minCornerNew%y
!    if (deltaNew%x <= 0.0 .or. deltaNew%y <= 0.0) then
!      status = -1
!  return
!    end if

    if (all(oldImage .eq. 0.0)) then
      status = ibset(status, 0)
  return
    end if

    if  (minval(oldPixelCorners%x) >= maxCornerNew%x&
    .or. maxval(oldPixelCorners%x) <= minCornerNew%x&
    .or. minval(oldPixelCorners%y) >= maxCornerNew%y&
    .or. maxval(oldPixelCorners%y) <= minCornerNew%y) then
      status = ibset(status, 1)
  return
    end if

    newImage = 0.0

    ! Check to see whether poly vertices are in clockwise or
    ! anticlockwise order:
    poly(1)%pos = oldPixelCorners(1, 1)
    poly(2)%pos = oldPixelCorners(1, 2)
    poly(3)%pos = oldPixelCorners(2, 2)
    poly(4)%pos = oldPixelCorners(2, 1)
    totalPolyArea = areaOfPoly(poly%pos)
    if (totalPolyArea > 0.0) then
      parity = 1
    else
      parity = -1
    end if

!print *, '  oldXsize, oldYsize:', oldXsize, oldYsize

    ! Starting point: (i) a real array of size (oldXsize, oldYsize) that
    ! contains the old image values; (ii) an array of type pointType and
    ! size (oldXsize+1, oldYsize+1) which contains the xy values of each
    ! pixel corner of the old image in the new coord system; (iii) an empty
    ! real array of size (newXsize, newYsize) which will contain the
    ! regridded image, with associated minXY and maxXY values.

    if (parity > 0.0) then
      ! Process each old image pixel:
      do oldXi = 1, oldXsize
        do oldYi = 1, oldYsize
        if (oldImage(oldXi, oldYi) .eq. 0.0) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          poly(1)%pos = oldPixelCorners(oldXi,     oldYi)
          poly(2)%pos = oldPixelCorners(oldXi,     oldYi + 1)
          poly(3)%pos = oldPixelCorners(oldXi + 1, oldYi + 1)
          poly(4)%pos = oldPixelCorners(oldXi + 1, oldYi)

          if  (minval(poly%pos%x) .ge. maxCornerNew%x&
          .or. maxval(poly%pos%x) .le. minCornerNew%x&
          .or. minval(poly%pos%y) .ge. maxCornerNew%y&
          .or. maxval(poly%pos%y) .le. minCornerNew%y)&
        cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid(poly, oldImage(oldXi, oldYi)&
            , minCornerNew, maxCornerNew, newImage, foreignStatus)
!, deltaNew

!*** absorb foreignStatus into status
          if (foreignStatus .lt. 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
            status = -2
  return
          else if (foreignStatus .gt. 0) then
!            status = status + foreignStatus * 8
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if
        end do
      end do

    else ! parity < 0
      ! Process each old image pixel:
      do oldXi = 1, oldXsize
        do oldYi = 1, oldYsize
!oldXi = 23
!oldYi = 23
        if (oldImage(oldXi, oldYi) .eq. 0.0) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          poly(4)%pos = oldPixelCorners(oldXi,     oldYi)
          poly(3)%pos = oldPixelCorners(oldXi,     oldYi + 1)
          poly(2)%pos = oldPixelCorners(oldXi + 1, oldYi + 1)
          poly(1)%pos = oldPixelCorners(oldXi + 1, oldYi)

!          if  (minval(poly%pos%x) .ge. maxNew%x&
!          .or. maxval(poly%pos%x) .le. minNew%x&
!          .or. minval(poly%pos%y) .ge. maxNew%y&
!          .or. maxval(poly%pos%y) .le. minNew%y)&
!        cycle
  if  (minval(poly%pos%x) >= maxCornerNew%x&
  .or. maxval(poly%pos%x) <= minCornerNew%x&
  .or. minval(poly%pos%y) >= maxCornerNew%y&
  .or. maxval(poly%pos%y) <= minCornerNew%y)&
cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid(poly, oldImage(oldXi, oldYi)&
            , minCornerNew, maxCornerNew, newImage, foreignStatus)
!, deltaNew

!*** check foreignStatus into status
          if (foreignStatus .lt. 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
            status = -2
  return
          else if (foreignStatus .gt. 0) then
!            status = status + foreignStatus * 8
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if

        end do
      end do
    end if

  end subroutine regrid

  !---------------------------------------------------------------------
  subroutine calcIntersectPolyGrid(poly, value, minCornerNew, maxCornerNew&
  , newImage, status)
!, deltaNew
  !---------------------------------------------------------------------
  ! This sub takes a polygonal bin (perhaps, but not necessarily, a
  ! distorted image pixel) which contains a real value and divides the
  ! value up (on a per area basis) between the pixels of the array
  ! newImage(). Although the indices of newImage() extend between 1
  ! and size(newImage, 1) or size(newImage, 2), only pixels with x index
  ! between lowNewXi and hiNewXi and y index between lowNewYi and hiNewYi
  ! are added to. The left and right world-coordinate edges of the
  ! newImage() array are stored in minNew and maxNew.

    implicit none

    type(localPolyType), intent(in) :: poly(:)
    type(PointType), intent(in) :: minCornerNew, maxCornerNew
!, deltaNew
    real, intent(inout) :: newImage(:,:)
    real, intent(in) :: value
    integer, intent(out) :: status

    type(localPolyType) ::&
      slicePolys(size(poly), 2*size(poly)), &
      chunkPolys(size(poly), 2*size(poly))
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.

    integer ::&
      xi, yi, numSlicePolyPoints(size(poly)), numSlicePolys,&
      numChunkPolys, sliceStatus, numSlicePoly, numChunkPoly,&
      numChunkPolyPoints(size(poly)),&
      loNewXi, hiNewXi, loNewYi, hiNewYi, newXsize, newYsize
    real :: totalPolyArea, realNumCols, realNumRows, xLo, xHi, yLo, yHi, area

    !-------------------------------------------------------------------
    ! Start

!print *, '    started calcIntersectPolyGrid'
!print *, '    value=', value
    status = 0

    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)
    realNumCols = real(newXsize)
    realNumRows = real(newYsize)

    ! Calculate minigrid limits:
    loNewXi = max(nint(minval(poly%pos%x)), 1)
    hiNewXi = min(nint(maxval(poly%pos%x)), newXsize)
    loNewYi = max(nint(minval(poly%pos%y)), 1)
    hiNewYi = min(nint(maxval(poly%pos%y)), newYsize)

!print *, '    poly(1)%pos%x, poly(1)%pos%y:', poly(1)%pos%x, poly(1)%pos%y
!print *, '    poly(2)%pos%x, poly(2)%pos%y:', poly(2)%pos%x, poly(2)%pos%y
!print *, '    poly(3)%pos%x, poly(3)%pos%y:', poly(3)%pos%x, poly(3)%pos%y
!print *, '    poly(4)%pos%x, poly(4)%pos%y:', poly(4)%pos%x, poly(4)%pos%y
!print *, '    loNewXi, hiNewXi:', loNewXi, hiNewXi
!print *, '    loNewYi, hiNewYi:', loNewYi, hiNewYi
!print *, '    newImage(loNewXi,loNewYi) before:', newImage(loNewXi,loNewYi)
!print *, ''
    ! Make sure poly vertices are in clockwise order:
    totalPolyArea = areaOfPoly(poly%pos)
    if (totalPolyArea .lt. 0.0) then
      status = -1
  return
    end if

    ! Slice up the polygon vertically:
    do xi = loNewXi, hiNewXi
!print *, '      xi=:', xi
      xLo = fracVal(minCornerNew%x, maxCornerNew%x, (xi - 1) / realNumCols)
      xHi = fracVal(minCornerNew%x, maxCornerNew%x,  xi      / realNumCols)
!print *, '      xLo, xHi =:', xLo, xHi

      call verticalSlice(&
        poly,&
        xLo, xHi, slicePolys, &
        numSlicePolys, numSlicePolyPoints, sliceStatus&
      )

      if (sliceStatus .lt. 0) then
        status = -1
  return
      end if

      if (numSlicePolys .gt. 1) status = ibset(status, 0)

    if (numSlicePolys .le. 0) cycle

      ! Slice up the polygon horizontally:
      do yi = loNewYi, hiNewYi
!print *, '        yi=:', yi
        yLo = fracVal(minCornerNew%y, maxCornerNew%y, (yi - 1) / realNumRows)
        yHi = fracVal(minCornerNew%y, maxCornerNew%y,  yi      / realNumRows)
!print *, '        yLo, yHi =:', yLo, yHi
        do numSlicePoly = 1, numSlicePolys
!print *, '          numSlicePoly=:', numSlicePoly
          call horizontalSlice(&
            slicePolys(numSlicePoly, 1:numSlicePolyPoints(numSlicePoly)), &
            yLo, yHi, chunkPolys, &
            numChunkPolys, numChunkPolyPoints, sliceStatus&
          )

          if (sliceStatus .lt. 0) then
            status = -2
  return
          end if

          if (numChunkPolys .gt. 1) status = ibset(status, 1)

        if (numChunkPolys .le. 0) cycle

          do numChunkPoly = 1, numChunkPolys
            area = areaOfPoly(&
              chunkPolys(numChunkPoly, 1:numChunkPolyPoints(numChunkPoly))%pos&
            )
            newImage(xi, yi) = newImage(xi, yi)&
              + area * value / totalPolyArea
          end do ! loop over sub-polys in chunk
        end do ! loop over sub-polys in slice
      end do ! loop over horizontal bands
    end do ! loop over vertical bands
!print *, '    newImage(loNewXi,loNewYi) after:', newImage(loNewXi,loNewYi)

  end subroutine calcIntersectPolyGrid

  !---------------------------------------------------------------------
  subroutine regrid_20030401(oldPixelCorners, oldImage, newImage&
  , startNewXi, startNewYi&
  , status)
  !---------------------------------------------------------------------
  ! This sub is intended to allow rebinning of an image from one coordinate
  ! system to another. Now, we define here an image as a two-dimensional
  ! array of uniform rectangular pixels. If we change the coordinate system,
  ! this image becomes distorted. It is nice to be able to convert it once
  ! again to an array of rectangular pixels, but this time in the new
  ! coordinate system. This involves taking each of the old, distorted
  ! pixels and dividing its contents up among the new pixels. It is assumed
  ! here that (i) the distorted pixel still has straight 'sides', ie it is
  ! a quadrilateral (a polygon for generality); (ii) that the value within
  ! the old pixel is evenly distributed through the pixel (this assumption
  ! MAY BE INVALID for extremely nonlinear distortions); (iii) that the
  ! distorted pixel is not folded over - ie that no two of its sides cross.
  !
  ! The relation between the old and new coordinate systems is here entirely contained within the input array oldPixelCorners. This gives the coordinates of each pixel in the array oldImage, expressed in the wcs 'pixel' system of newImage. What does 'wcs pixel system' mean? It means that the width and height of pixels in newImage are both equal to 1.0 and the centre of the pixel newImage(1,1) is at (1.0, 1.0).
  !---------------------------------------------------------------------

    implicit none

    real,           intent(in)            :: oldImage(:,:)
    type(pointType),intent(in)            :: oldPixelCorners(size(oldImage,1)+1&
                                                            ,size(oldImage,2)+1)
    integer,        intent(in),  optional :: startNewXi,&
                                             startNewYi
    real,           intent(out)           :: newImage(:,:)
    integer,        intent(out), optional :: status

!    type(pointType) ::&
!      minNew, maxNew
    type(localPolyType) :: poly(4)
    integer :: finisNewXi, finisNewYi, parity
    integer :: oldXi, oldXsize, oldYi, oldYsize,&
!      lowNewXi, hiNewXi, lowNewYi, hiNewYi,&
      newXsize, newYsize, foreignStatus
    type(pointType) :: minCornerNew, maxCornerNew, deltaNew
!    logical :: xIsInverted, yIsInverted
    real :: totalPolyArea
!integer, pointer :: xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    oldXsize = size(oldImage, 1)
    oldYsize = size(oldImage, 2)
    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)

    if (newXsize .le. 0 .or. newYsize .le. 0) then
      status = ibset(status, 2)
  return
    end if

    if (present(startNewXi)) then
      finisNewXi = startNewXi + newXsize - 1
      minCornerNew%x = startNewXi - 0.5
    else
      finisNewXi = newXsize
      minCornerNew%x = - 0.5
    end if
    if (present(startNewYi)) then
      finisNewYi = startNewYi + newYsize - 1
      minCornerNew%y = startNewYi - 0.5
    else
      finisNewYi = newYsize
      minCornerNew%y = - 0.5
    end if
    maxCornerNew%x = finisNewXi + 0.5
    maxCornerNew%y = finisNewYi + 0.5

    deltaNew%x = maxCornerNew%x - minCornerNew%x
    deltaNew%y = maxCornerNew%y - minCornerNew%y
    if (deltaNew%x <= 0.0 .or. deltaNew%y <= 0.0) then
      status = -1
  return
    end if

    if (all(oldImage .eq. 0.0)) then
      status = ibset(status, 0)
  return
    end if

    if  (minval(oldPixelCorners%x) >= maxCornerNew%x&
    .or. maxval(oldPixelCorners%x) <= minCornerNew%x&
    .or. minval(oldPixelCorners%y) >= maxCornerNew%y&
    .or. maxval(oldPixelCorners%y) <= minCornerNew%y) then
      status = ibset(status, 1)
  return
    end if

    newImage = 0.0

    ! Check to see whether poly vertices are in clockwise or
    ! anticlockwise order:
    poly(1)%pos = oldPixelCorners(1, 1)
    poly(2)%pos = oldPixelCorners(1, 2)
    poly(3)%pos = oldPixelCorners(2, 2)
    poly(4)%pos = oldPixelCorners(2, 1)
    totalPolyArea = areaOfPoly(poly%pos)
    if (totalPolyArea > 0.0) then
      parity = 1
    else
      parity = -1
    end if

print *, '  oldXsize, oldYsize:', oldXsize, oldYsize

!print *, '  parity:'&
!, parity
!print *, '  oldPixelCorners(1,1)%x, oldPixelCorners(1,1)%y:'&
!, oldPixelCorners(1,1)%x, oldPixelCorners(1,1)%y
!print *, '  oldPixelCorners(oldXsize+1,1)%x, oldPixelCorners(oldXsize+1,1)%y:'&
!, oldPixelCorners(oldXsize+1,1)%x, oldPixelCorners(oldXsize+1,1)%y
!print *, '  oldPixelCorners(1,oldYsize+1)%x, oldPixelCorners(1,oldYsize+1)%y:'&
!, oldPixelCorners(1,oldYsize+1)%x, oldPixelCorners(1,oldYsize+1)%y
!print *, '  oldPixelCorners(oldXsize+1,oldYsize+1)%x, oldPixelCorners(oldXsize+1,oldYsize+1)%y:'&
!, oldPixelCorners(oldXsize+1,oldYsize+1)%x, oldPixelCorners(oldXsize+1,oldYsize+1)%y
!print *, '  minCornerNew%x, maxCornerNew%x:'&
!, minCornerNew%x, maxCornerNew%x
!print *, '  minCornerNew%y, maxCornerNew%y:'&
!, minCornerNew%y, maxCornerNew%y

    ! Starting point: (i) a real array of size (oldXsize, oldYsize) that
    ! contains the old image values; (ii) an array of type pointType and
    ! size (oldXsize+1, oldYsize+1) which contains the xy values of each
    ! pixel corner of the old image in the new coord system; (iii) an empty
    ! real array of size (newXsize, newYsize) which will contain the
    ! regridded image, with associated minXY and maxXY values.

    if (parity > 0.0) then
      ! Process each old image pixel:
      do oldXi = 1, oldXsize
        do oldYi = 1, oldYsize
        if (oldImage(oldXi, oldYi) .eq. 0.0) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          poly(1)%pos = oldPixelCorners(oldXi,     oldYi)
          poly(2)%pos = oldPixelCorners(oldXi,     oldYi + 1)
          poly(3)%pos = oldPixelCorners(oldXi + 1, oldYi + 1)
          poly(4)%pos = oldPixelCorners(oldXi + 1, oldYi)

          if  (minval(poly%pos%x) .ge. maxCornerNew%x&
          .or. maxval(poly%pos%x) .le. minCornerNew%x&
          .or. minval(poly%pos%y) .ge. maxCornerNew%y&
          .or. maxval(poly%pos%y) .le. minCornerNew%y)&
        cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid_20030401(poly, oldImage(oldXi, oldYi)&
            , minCornerNew, maxCornerNew, deltaNew, newImage, foreignStatus)

!*** absorb foreignStatus into status
          if (foreignStatus .lt. 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
            status = -2
  return
          else if (foreignStatus .gt. 0) then
!            status = status + foreignStatus * 8
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if
        end do
      end do

    else ! parity < 0
      ! Process each old image pixel:
!!!      do oldXi = 1, oldXsize
!!!        do oldYi = 1, oldYsize
oldXi = 23
oldYi = 23
!!!        if (oldImage(oldXi, oldYi) .eq. 0.0) cycle

          ! Load the coordinates of the pixel corners into a polygon:
          poly(4)%pos = oldPixelCorners(oldXi,     oldYi)
          poly(3)%pos = oldPixelCorners(oldXi,     oldYi + 1)
          poly(2)%pos = oldPixelCorners(oldXi + 1, oldYi + 1)
          poly(1)%pos = oldPixelCorners(oldXi + 1, oldYi)

!          if  (minval(poly%pos%x) .ge. maxNew%x&
!          .or. maxval(poly%pos%x) .le. minNew%x&
!          .or. minval(poly%pos%y) .ge. maxNew%y&
!          .or. maxval(poly%pos%y) .le. minNew%y)&
!        cycle
!!!  if  (minval(poly%pos%x) >= maxCornerNew%x&
!!!  .or. maxval(poly%pos%x) <= minCornerNew%x&
!!!  .or. minval(poly%pos%y) >= maxCornerNew%y&
!!!  .or. maxval(poly%pos%y) <= minCornerNew%y)&
!!!cycle

!*** check for no lines crossing??

          call calcIntersectPolyGrid_20030401(poly, oldImage(oldXi, oldYi)&
            , minCornerNew, maxCornerNew, deltaNew, newImage, foreignStatus)

!*** check foreignStatus into status
          if (foreignStatus .lt. 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
            status = -2
  return
          else if (foreignStatus .gt. 0) then
!            status = status + foreignStatus * 8
            status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
          end if

!!!        end do
!!!      end do
    end if

  end subroutine regrid_20030401

  !---------------------------------------------------------------------
  subroutine calcIntersectPolyGrid_20030401(poly, value&
  , minCornerNew, maxCornerNew, deltaNew, newImage, status)
  !---------------------------------------------------------------------
  ! This sub takes a polygonal bin (perhaps, but not necessarily, a
  ! distorted image pixel) which contains a real value and divides the
  ! value up (on a per area basis) between the pixels of the array
  ! newImage(). Although the indices of newImage() extend between 1
  ! and size(newImage, 1) or size(newImage, 2), only pixels with x index
  ! between lowNewXi and hiNewXi and y index between lowNewYi and hiNewYi
  ! are added to. The left and right world-coordinate edges of the
  ! newImage() array are stored in minNew and maxNew.

    implicit none

    type(localPolyType), intent(in) :: poly(:)
    type(PointType), intent(in) :: minCornerNew, maxCornerNew, deltaNew
    real, intent(inout) :: newImage(:,:)
    real, intent(in) :: value
    integer, intent(out) :: status

    type(localPolyType) ::&
      slicePolys(size(poly), 2*size(poly)), &
      chunkPolys(size(poly), 2*size(poly))
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.

    integer ::&
      xi, yi, numSlicePolyPoints(size(poly)), numSlicePolys,&
      numChunkPolys, sliceStatus, numSlicePoly, numChunkPoly,&
      numChunkPolyPoints(size(poly)),&
      lowNewXi, hiNewXi, lowNewYi, hiNewYi, newXsize, newYsize
    real :: totalPolyArea, realNumCols, realNumRows, xLo, xHi, yLo, yHi, area

    !-------------------------------------------------------------------
    ! Start

print *, '    started calcIntersectPolyGrid'
    status = 0

    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)
    realNumCols = real(newXsize)
    realNumRows = real(newYsize)

    ! Calculate minigrid limits:
    lowNewXi = max(1 + floor(newXsize * ((minval(poly%pos%x) - minCornerNew%x)&
      / deltaNew%x)), 1)
    hiNewXi  = min(  ceiling(newXsize * ((maxval(poly%pos%x) - minCornerNew%x)&
      / deltaNew%x)), newXsize)

    lowNewYi = max(1 + floor(newYsize * ((minval(poly%pos%y) - minCornerNew%y)&
      / deltaNew%y)), 1)
    hiNewYi  = min(  ceiling(newYsize * ((maxval(poly%pos%y) - minCornerNew%y)&
      / deltaNew%y)), newYsize)
print *, '    poly(1)%pos%x, poly(1)%pos%y:', poly(1)%pos%x, poly(1)%pos%y
print *, '    poly(2)%pos%x, poly(2)%pos%y:', poly(2)%pos%x, poly(2)%pos%y
print *, '    poly(3)%pos%x, poly(3)%pos%y:', poly(3)%pos%x, poly(3)%pos%y
print *, '    poly(4)%pos%x, poly(4)%pos%y:', poly(4)%pos%x, poly(4)%pos%y
print *, '    lowNewXi, hiNewXi:', lowNewXi, hiNewXi
print *, '    lowNewYi, hiNewYi:', lowNewYi, hiNewYi
print *, '    newImage(lowNewXi,lowNewYi) before:', newImage(lowNewXi,lowNewYi)
    ! Make sure poly vertices are in clockwise order:
    totalPolyArea = areaOfPoly(poly%pos)
    if (totalPolyArea .lt. 0.0) then
      status = -1
  return
    end if

    ! Slice up the polygon vertically:
    do xi = lowNewXi, hiNewXi
      xLo = fracVal(minCornerNew%x, maxCornerNew%x, (xi - 1) / realNumCols)
      xHi = fracVal(minCornerNew%x, maxCornerNew%x,  xi      / realNumCols)

      call verticalSlice(&
        poly,&
        xLo, xHi, slicePolys, &
        numSlicePolys, numSlicePolyPoints, sliceStatus&
      )

      if (sliceStatus .lt. 0) then
        status = -1
  return
      end if

      if (numSlicePolys .gt. 1) status = ibset(status, 0)

    if (numSlicePolys .le. 0) cycle

      ! Slice up the polygon horizontally:
      do yi = lowNewYi, hiNewYi
        yLo = fracVal(minCornerNew%y, maxCornerNew%y, (yi - 1) / realNumRows)
        yHi = fracVal(minCornerNew%y, maxCornerNew%y,  yi      / realNumRows)
        do numSlicePoly = 1, numSlicePolys
          call horizontalSlice(&
            slicePolys(numSlicePoly, 1:numSlicePolyPoints(numSlicePoly)), &
            yLo, yHi, chunkPolys, &
            numChunkPolys, numChunkPolyPoints, sliceStatus&
          )

          if (sliceStatus .lt. 0) then
            status = -2
  return
          end if

          if (numChunkPolys .gt. 1) status = ibset(status, 1)

        if (numChunkPolys .le. 0) cycle

          do numChunkPoly = 1, numChunkPolys
            area = areaOfPoly(&
              chunkPolys(numChunkPoly, 1:numChunkPolyPoints(numChunkPoly))%pos&
            )
            newImage(xi, yi) = newImage(xi, yi)&
              + area * value / totalPolyArea
          end do ! loop over sub-polys in chunk
        end do ! loop over sub-polys in slice
      end do ! loop over horizontal bands
    end do ! loop over vertical bands
print *, '    newImage(lowNewXi,lowNewYi) after:', newImage(lowNewXi,lowNewYi)

  end subroutine calcIntersectPolyGrid_20030401

  !---------------------------------------------------------------------
  subroutine regrid_20021030(oldPixelCorners, oldImage, newImage, minNewArg&
  , maxNewArg, status)
  !---------------------------------------------------------------------
  ! This sub is intended to allow rebinning of an image from one coordinate
  ! system to another. Now, we define here an image as a two-dimensional
  ! array of uniform rectangular pixels. If we change the coordinate system,
  ! this image becomes distorted. It is nice to be able to convert it once
  ! again to an array of rectangular pixels, but this time in the new
  ! coordinate system. This involves taking each of the old, distorted
  ! pixels and dividing its contents up among the new pixels. It is assumed
  ! here that (i) the distorted pixel still has straight 'sides', ie it is
  ! a quadrilateral (a polygon for generality); (ii) that the value within
  ! the old pixel is evenly distributed through the pixel (this assumption
  ! MAY BE INVALID for extremely nonlinear distortions); (iii) that the
  ! distorted pixel is not folded over - ie that no two of its sides cross.
  !

!    use sscmisc

    implicit none

    real, intent(in) ::&
      oldImage(:,:)
    type(pointType), intent(in) ::&
      oldPixelCorners(size(oldImage, 1) + 1, size(oldImage, 2) + 1),&
      minNewArg, maxNewArg
    real, intent(out) :: newImage(:,:)
    integer, intent(out) :: status

    type(pointType) ::&
      minNew, maxNew
    type(localPolyType) :: poly(4)
    integer :: oldXi, oldXsize, oldYi, oldYsize, lowNewXi, hiNewXi,&
      lowNewYi, hiNewYi, newXsize, newYsize, foreignStatus
    type(pointType) :: deltaNew
    logical :: xIsInverted, yIsInverted
    real :: totalPolyArea
integer, pointer :: xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    minNew = minNewArg
    maxNew = maxNewArg

    deltaNew%x = maxNew%x - minNew%x
    deltaNew%y = maxNew%y - minNew%y
    if (deltaNew%x .eq. 0.0 .or. deltaNew%y .eq. 0.0) then
      status = -1
  return
    end if

    if (deltaNew%x .lt. 0.0) then
      xIsInverted = .true.
      deltaNew%x = abs(deltaNew%x)
      call swap(maxNew%x, minNew%x)
    else
      xIsInverted = .false.
    end if
    if (deltaNew%y .lt. 0.0) then
      yIsInverted = .true.
      deltaNew%y = abs(deltaNew%y)
      call swap(maxNew%y, minNew%y)
    else
      yIsInverted = .false.
    end if

    if (all(oldImage .eq. 0.0)) then
      status = ibset(status, 0)
  return
    end if

    if  (minval(oldPixelCorners%x) .ge. maxNew%x&
    .or. maxval(oldPixelCorners%x) .le. minNew%x&
    .or. minval(oldPixelCorners%y) .ge. maxNew%y&
    .or. maxval(oldPixelCorners%y) .le. minNew%y) then
      status = ibset(status, 1)
  return
    end if

    oldXsize = size(oldImage, 1)
    oldYsize = size(oldImage, 2)
    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)

    if (newXsize .le. 0 .or. newYsize .le. 0) then
      status = ibset(status, 2)
  return
    end if

    newImage = 0.0

    ! Starting point: (i) an real array of size (oldXsize, oldYsize) that
    ! contains the old image values; (ii) an array of type pointType and
    ! size (oldXsize+1, oldYsize+1) which contains the xy values of each
    ! pixel corner of the old image in the new coord system; (iii) an empty
    ! real array of size (newXsize, newYsize) which will contain the
    ! regridded image, with associated minXY and maxXY values.

    ! Process each old image pixel:
    do oldXi = 1, oldXsize
      do oldYi = 1, oldYsize
      if (oldImage(oldXi, oldYi) .eq. 0.0) cycle

        ! Load the coordinates of the pixel corners into a polygon:
        poly(1)%pos = oldPixelCorners(oldXi,     oldYi)
        poly(2)%pos = oldPixelCorners(oldXi,     oldYi + 1)
        poly(3)%pos = oldPixelCorners(oldXi + 1, oldYi + 1)
        poly(4)%pos = oldPixelCorners(oldXi + 1, oldYi)

        if  (minval(poly%pos%x) .ge. maxNew%x&
        .or. maxval(poly%pos%x) .le. minNew%x&
        .or. minval(poly%pos%y) .ge. maxNew%y&
        .or. maxval(poly%pos%y) .le. minNew%y)&
      cycle

        ! Calculate minigrid limits:
        lowNewXi = max(1 + floor(newXsize * ((minval(poly%pos%x) - minNew%x)&
        / deltaNew%x)), 1)
        hiNewXi  = min(  ceiling(newXsize * ((maxval(poly%pos%x) - minNew%x)&
        / deltaNew%x)), newXsize)

        lowNewYi = max(1 + floor(newYsize * ((minval(poly%pos%y) - minNew%y)&
        / deltaNew%y)), 1)
        hiNewYi  = min(  ceiling(newYsize * ((maxval(poly%pos%y) - minNew%y)&
        / deltaNew%y)), newYsize)

!*** check for no lines crossing??

        ! Make sure poly vertices are in clockwise order:
        totalPolyArea = areaOfPoly(poly%pos)
        if (totalPolyArea .lt. 0.0) then
          status = -2
  return
        end if

        call calcIntersectPolyGrid_20021030(poly, totalPolyArea&
        , oldImage(oldXi, oldYi)&
        , lowNewXi, hiNewXi, lowNewYi, hiNewYi, minNew, maxNew&
        , newImage, foreignStatus)
!*** absorb foreignStatus into status
        if (foreignStatus .lt. 0) then
print *, 'calcIntersectPolyGrid err code =', foreignStatus
          status = -3
  return
        else if (foreignStatus .gt. 0) then
!          status = status + foreignStatus * 8
          status = ior(status, ishft(foreignStatus, 3))
!***** or -3???
        end if
      end do
    end do

    if (xIsInverted) then
!      newImages = newImages(:, (newXsize + 1 - indicize(newXsize)), :)
xi => indicize(newXsize)
newImage = newImage((newXsize + 1 - xi), :)
deallocate(xi)
    end if
    if (yIsInverted) then
!      newImages = newImages(:, :, (newYsize + 1 - indicize(newYsize)))
yi => indicize(newYsize)
newImage = newImage(:, (newYsize + 1 - yi))
deallocate(yi)
    end if

  end subroutine regrid_20021030

  !---------------------------------------------------------------------
  subroutine calcIntersectPolyGrid_20021030(poly, totalPolyArea, value&
  , lowNewXi, hiNewXi, lowNewYi, hiNewYi, minNew, maxNew, newImage, status)
  !---------------------------------------------------------------------
  ! This sub takes a polygonal bin (perhaps, but not necessarily, a
  ! distorted image pixel) which contains a real value and divides the
  ! value up (on a per area basis) between the pixels of the array
  ! newImage(). Although the indices of newImage() extend between 1
  ! and size(newImage, 1) or size(newImage, 2), only pixels with x index
  ! between lowNewXi and hiNewXi and y index between lowNewYi and hiNewYi
  ! are added to. The left and right world-coordinate edges of the
  ! newImage() array are stored in minNew and maxNew.

!    use mathutils_mod

    implicit none

    type(localPolyType), intent(in) :: poly(:)
    real, intent(inout) :: newImage(:,:)
    real, intent(in) :: totalPolyArea, value
    integer, intent(in) :: lowNewXi, hiNewXi, lowNewYi, hiNewYi
    type(pointType), intent(in) :: minNew, maxNew
    integer, intent(out) :: status

    type(localPolyType) ::&
      slicePolys(size(poly), 2*size(poly)), &
      chunkPolys(size(poly), 2*size(poly))
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.

    integer ::&
      xi, yi, numSlicePolyPoints(size(poly)), numSlicePolys,&
      numChunkPolys, sliceStatus, numSlicePoly, numChunkPoly,&
      numChunkPolyPoints(size(poly)), newXsize, newYsize
    real :: realNumCols, realNumRows, xLo, xHi, yLo, yHi, area

    !-------------------------------------------------------------------
    ! Start

    status = 0

    newXsize = size(newImage, 1)
    newYsize = size(newImage, 2)
    realNumCols = real(newXsize)
    realNumRows = real(newYsize)

    ! Slice up the polygon vertically:
    do xi = lowNewXi, hiNewXi
      xLo = fracVal(minNew%x, maxNew%x, (xi - 1) / realNumCols)
      xHi = fracVal(minNew%x, maxNew%x,  xi      / realNumCols)

      call verticalSlice(&
        poly,&
        xLo, xHi, slicePolys, &
        numSlicePolys, numSlicePolyPoints, sliceStatus&
      )

      if (sliceStatus .lt. 0) then
        status = -1
  return
      end if

      if (numSlicePolys .gt. 1) status = ibset(status, 0)

    if (numSlicePolys .le. 0) cycle

      ! Slice up the polygon horizontally:
      do yi = lowNewYi, hiNewYi
        yLo = fracVal(minNew%y, maxNew%y, (yi - 1) / realNumRows)
        yHi = fracVal(minNew%y, maxNew%y,  yi      / realNumRows)
        do numSlicePoly = 1, numSlicePolys
          call horizontalSlice(&
            slicePolys(numSlicePoly, 1:numSlicePolyPoints(numSlicePoly)), &
            yLo, yHi, chunkPolys, &
            numChunkPolys, numChunkPolyPoints, sliceStatus&
          )

          if (sliceStatus .lt. 0) then
            status = -2
  return
          end if

          if (numChunkPolys .gt. 1) status = ibset(status, 1)

        if (numChunkPolys .le. 0) cycle

          do numChunkPoly = 1, numChunkPolys
            area = areaOfPoly(&
              chunkPolys(numChunkPoly, 1:numChunkPolyPoints(numChunkPoly))%pos&
            )
            newImage(xi, yi) = newImage(xi, yi)&
              + area * value / totalPolyArea
          end do ! loop over sub-polys in chunk
        end do ! loop over sub-polys in slice
      end do ! loop over horizontal bands
    end do ! loop over vertical bands

  end subroutine calcIntersectPolyGrid_20021030

  !---------------------------------------------------------------------
  subroutine verticalSlice(poly, xLo, xHi, subPolys&
  , nSubs, nPoints, status)
  !---------------------------------------------------------------------
  ! NOTE that it is ABSOLUTELY REQUIRED in this subroutine that the points
  ! in the poly are numbered in clockwise order and that none of the sides
  ! cross each other.
  !
  ! The sub has been copied to horizontalSlice with minimal alterations.
  ! Therefore if bugs are found in one, the other should also be altered.

!    use mathutils_mod

    implicit none

    type(localPolyType), intent(in) :: poly(:)
    real, intent(in) :: xLo, xHi
    type(localPolyType), intent(out) :: subPolys(size(poly),2*size(poly))
    ! The dimensions given are the maximum possible limits. The first index
    ! is the number of the sub-poly resulting from the intersection of poly
    ! with the two verticals, the second index is the number of the point
    ! within that sub-poly.
    integer, intent(out) :: nPoints(size(poly)), nSubs, status

    type :: intersectionType
      real :: y
      logical :: exists, isAvailable
    end type intersectionType
    type(intersectionType) :: interLo(size(poly)), interHi(size(poly))
    integer :: i, j, k, zone(size(poly)), nVertices, precedingVertex,&
      state, indexVec(1)
    integer, parameter :: LOW = 1, HIGH = 2, MIDDLE = 3
    real :: xFrac
    logical :: vertexIsAvailable(size(poly)), mask(size(poly))
logical :: flag = .false.

    !-------------------------------------------------------------------
    ! Start

if (flag) then
print *, '      -----------------------------------------------------------'
print *, '      Now inside verticalSlice.'
print *, ''
end if

    status = 0
    nPoints = 0 ! vector default
    nSubs = 0 ! default

if (flag) then
print *, '      minval(poly%pos%x):', minval(poly%pos%x)
print *, '      maxval(poly%pos%x):', maxval(poly%pos%x)
print *, '                     xLo:', xLo
print *, '                     xHi:', xHi
end if

  if (minval(poly%pos%x) .ge. xHi .or. maxval(poly%pos%x) .le. xLo) return

    nVertices = size(poly)

if (flag) then
print *, '               nVertices:', nVertices
print *, ''
end if

! Find out into which zone each poly vertex falls:

    do i = 1, nVertices
      if (poly(i)%pos%x .le. xLo) then
        zone(i) = LOW
      else if (poly(i)%pos%x .ge. xHi) then
        zone(i) = HIGH
      else
        zone(i) = MIDDLE
      end if
    end do

! Find which sides of the polygon intersect the left and right verticals:

    interLo%isAvailable = .false.
    interHi%isAvailable = .false.
    interLo%exists = .false.
    interHi%exists = .false.
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices) ! ie j=1 for i=nVertices, j=i+1 else.

      if ( (poly(i)%pos%x .le. xLo .and. poly(j)%pos%x .gt. xLo)&
      .or. (poly(i)%pos%x .gt. xLo .and. poly(j)%pos%x .le. xLo)) then
        xFrac = (xLo - poly(i)%pos%x) / (poly(j)%pos%x - poly(i)%pos%x)
        interLo(i)%y = fracVal(poly(i)%pos%y, poly(j)%pos%y, xFrac)
        interLo(i)%exists      = .true.
        interLo(i)%isAvailable = .true.
      end if
      if ( (poly(i)%pos%x .ge. xHi .and. poly(j)%pos%x .lt. xHi)&
      .or. (poly(i)%pos%x .lt. xHi .and. poly(j)%pos%x .ge. xHi)) then
        xFrac = (xHi - poly(i)%pos%x) / (poly(j)%pos%x - poly(i)%pos%x)
        interHi(i)%y = fracVal(poly(i)%pos%y, poly(j)%pos%y, xFrac)
        interHi(i)%exists      = .true.
        interHi(i)%isAvailable = .true.
      end if
    end do

    ! And, just for neatness really, change a glancing touch onto one of the verticals from a non-MIDDLE point to a MIDDLE point. If this is not done, the intersection polygon will have 2 points at that place.
    k = 1 + modulo(1, nVertices)
    do i = 1, nVertices
      j = k				! ie j = i + 1, modulo nVertices
      k = 1 + modulo(i + 1, nVertices)	! ie k = i + 2, modulo nVertices
      if (zone(i) .eq. MIDDLE .and. zone(k) .eq. MIDDLE) then
        if (poly(j)%pos%x .eq. xLo) then
          if (isConvex(poly(i)%pos, poly(j)%pos, poly(k)%pos)) then
            zone(j) = MIDDLE
            interLo(i)%exists = .false.
            interLo(j)%exists = .false.
            interLo(i)%isAvailable = .false.
            interLo(j)%isAvailable = .false.
          end if
        else if (poly(j)%pos%x .eq. xHi) then
          if (isConvex(poly(i)%pos, poly(j)%pos, poly(k)%pos)) then
            zone(j) = MIDDLE
            interHi(i)%exists = .false.
            interHi(j)%exists = .false.
            interHi(i)%isAvailable = .false.
            interHi(j)%isAvailable = .false.
          end if
        end if
      end if
    end do

if (any(interLo%exists .and. interHi%exists .and. (zone .eq. MIDDLE))) then
!***** BUG!!
status = -1
return
end if

    vertexIsAvailable = .false. ! vector default
    where(zone .eq. MIDDLE) vertexIsAvailable = .true.

    ! Finally, because it is possible for poly()pos%x to be just on the other side of a vertical to its neighbouring vertices, thereby generating two neighbouring intercepts with that vertical, but these intercepts can nevertheless due to numerical rounding errors be at the same y value, the following is done to erase any such paired, same-y intercepts:
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices)	! ie j = i + 1, modulo nVertices

      if (interLo(i)%exists .and. interLo(j)%exists&
      .and. interLo(i)%y .eq. interLo(j)%y) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interLo(i)%exists = .false.
        interLo(j)%exists = .false.
        interLo(i)%isAvailable = .false.
        interLo(j)%isAvailable = .false.
if (flag) print *, 'taken out LO intercepts', i, 'and', j
      end if

      if (interHi(i)%exists .and. interHi(j)%exists&
      .and. interHi(i)%y .eq. interHi(j)%y) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interHi(i)%exists = .false.
        interHi(j)%exists = .false.
        interHi(i)%isAvailable = .false.
        interHi(j)%isAvailable = .false.
if (flag) print *, 'taken out HI intercepts', i, 'and', j
      end if
    end do

if (flag) then
do i = 1, nVertices
  print *, '                            i:', i
  print *, '                poly(i)%pos%x:', poly(i)%pos%x
  print *, '                poly(i)%pos%y:', poly(i)%pos%y
  print *, '                      zone(i):', zone(i)
  print *, '         vertexIsAvailable(i):', vertexIsAvailable(i)
  print *, '            interLo(i)%exists:', interLo(i)%exists
  print *, '       interLo(i)%isAvailable:', interLo(i)%isAvailable
  print *, '                 interLo(i)%y:', interLo(i)%y
  print *, '            interHi(i)%exists:', interHi(i)%exists
  print *, '       interHi(i)%isAvailable:', interHi(i)%isAvailable
  print *, '                 interHi(i)%y:', interHi(i)%y
  print *, ''
end do

print *, '      Ok, now constructing the intersection:'
print *, ''
end if

! Now construct the intersection of the verticals and the poly (could be a number of subpolys):

    do while(any(interLo%isAvailable) .or. any(interHi%isAvailable)&
    .or. any(vertexIsAvailable))

      ! Start a new subpoly:

      nSubs = nSubs + 1
      nPoints(nSubs) = 1

if (flag) then
print *, '      Start ',nSubs,'nth new sub-poly:'
print *, ''
end if
      if (any(interLo%isAvailable)) then ! find it:
        do i = 1, nVertices
          if (interLo(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interLo(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%pos%x = xLo
        subPolys(nSubs, nPoints(nSubs))%pos%y = interLo(precedingVertex)%y

        if (zone(precedingVertex) .eq. LOW) then
          state = MIDDLE
        else
          state = LOW
        end if

if (flag) then
print *, '      Ok, I found some available lo-x intercepts.'
print *, ''
print *, '      precedingVertex for the first is: ', precedingVertex
print *, ''
end if
      else if (any(interHi%isAvailable)) then
        do i = 1, nVertices
          if (interHi(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interHi(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%pos%x = xHi
        subPolys(nSubs, nPoints(nSubs))%pos%y = interHi(precedingVertex)%y

        if (zone(precedingVertex) .eq. HIGH) then
          state = MIDDLE
        else
          state = HIGH
        end if

if (flag) then
print *, '      Didnt find any available lo-x intercepts, but I found some hi-x intercepts available.'
print *, ''
print *, '      precedingVertex for the first is: ', precedingVertex
print *, ''
end if
      else ! there must be an available vertex.
        do i = 1, nVertices
          if (vertexIsAvailable(i)) then
            precedingVertex = i
        exit
          end if
        end do
        vertexIsAvailable(precedingVertex) = .false.
        subPolys(nSubs, nPoints(nSubs))%pos = poly(precedingVertex)%pos

        state = MIDDLE
if (flag) then
print *, '      Didnt find any available lo-x or hi-x intercepts, but I found some available polygon vertices.'
print *, ''
print *, '      Vertex of the first is: ', precedingVertex
print *, ''
end if
      end if

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'zone(precedingVertex=',precedingVertex,'): ', zone(precedingVertex)
print *, 'state: ', state
print *, ''
end if

      ! Now follow it around until the end:
      do
        select case(state)
        case(LOW)
if (flag) print *, 'state is LOW (=',LOW,').'
        ! Move up until next higher intercept, after which state ought to
        ! be MIDDLE.
          mask = (interLo%exists&
          .and. interLo%y .gt. subPolys(nSubs, nPoints(nSubs))%pos%y)
!          .and. interLo%y .ge. subPolys(nSubs, nPoints(nSubs))%pos%y)
if (.not. any(mask)) then
!***** BUG!!
status = -2
return
end if
          indexVec = minloc(interLo%y, mask)
          precedingVertex = indexVec(1)
if (flag) print *, 'precedingVertex is now: ', precedingVertex
      if (.not. interLo(precedingVertex)%isAvailable) exit ! because this
      ! subpoly is now finished.

if (zone(precedingVertex) .ne. LOW) then
!***** BUG!!
status = -3
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interLo(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%pos%x = xLo
          subPolys(nSubs, nPoints(nSubs))%pos%y = interLo(precedingVertex)%y

          state = MIDDLE

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

        case(HIGH)
if (flag) print *, 'state is HIGH (=',HIGH,').'
        ! Move down until next lower intercept, after which state ought to
        ! be MIDDLE.
          mask = (interHi%exists&
          .and. interHi%y .lt. subPolys(nSubs, nPoints(nSubs))%pos%y)
!          .and. interHi%y .le. subPolys(nSubs, nPoints(nSubs))%pos%y)
if (.not. any(mask)) then
!***** BUG!!
status = -4
return
end if
          indexVec = maxloc(interHi%y, mask)
          precedingVertex = indexVec(1)
if (flag) print *, 'precedingVertex is now: ', precedingVertex
      if (.not. interHi(precedingVertex)%isAvailable) exit ! because this
      ! subpoly is now finished.

if (zone(precedingVertex) .ne. HIGH) then
!***** BUG!!
status = -5
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interHi(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%pos%x = xHi
          subPolys(nSubs, nPoints(nSubs))%pos%y = interHi(precedingVertex)%y

          state = MIDDLE

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

        case(MIDDLE)
if (flag) print *, 'state is MIDDLE (=',MIDDLE,').'
          if (interLo(precedingVertex)%isAvailable) then
if (flag) print *, 'this side intersects xLo.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            interLo(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%pos%x = xLo
            subPolys(nSubs, nPoints(nSubs))%pos%y = interLo(precedingVertex)%y

            state = LOW

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: LOW'
print *, ''
end if

          else&
          if (interHi(precedingVertex)%isAvailable) then
if (flag) print *, 'this side intersects xHi.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            interHi(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%pos%x = xHi
            subPolys(nSubs, nPoints(nSubs))%pos%y = interHi(precedingVertex)%y

            state = HIGH

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: HIGH'
print *, ''
end if

          else
            precedingVertex = 1 + modulo(precedingVertex, nVertices)
if (flag) print *, 'vertex is now: ', precedingVertex
      if (.not. vertexIsAvailable(precedingVertex)) exit ! because this
      ! subpoly is now finished.

if (flag) print *, 'this side doesnt intersect either xLo or xHi.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            vertexIsAvailable(precedingVertex) = .false.
            subPolys(nSubs, nPoints(nSubs))%pos = poly(precedingVertex)%pos

!           state = MIDDLE ! - still.
if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

          end if

        case default
!***** BUG!!
status = -6
return
        end select
if (flag) print *, ''
      end do ! complete sub-poly
if (flag) print *, ''
    end do ! no more available vertices or intersections.
if (flag) print *, ''

  end subroutine verticalSlice

  !---------------------------------------------------------------------
  logical function isConvex(pos0, pos1, pos2)
  !---------------------------------------------------------------------

    implicit none

    type(pointType), intent(in) :: pos0, pos1, pos2

    type(pointType) :: vectorA, vectorB

    !-------------------------------------------------------------------
    ! Start

    if ( (pos0%x .eq. pos1%x .and. pos0%y .eq. pos1%y)&
    .or. (pos0%x .eq. pos2%x .and. pos0%y .eq. pos2%y)&
    .or. (pos1%x .eq. pos2%x .and. pos1%y .eq. pos2%y)) then
      isConvex = .true. ! why not? Neither answer is sensible in these cases.

    else
      vectorA%x = pos1%x - pos0%x
      vectorA%y = pos1%y - pos0%y
      vectorB%x = pos2%x - pos1%x
      vectorB%y = pos2%y - pos1%y
      if (vectorA%x * vectorB%y .le. vectorA%y * vectorB%x) then
        isConvex = .true.
      else
        isConvex = .false.
      end if
    end if

  end function isConvex

  !---------------------------------------------------------------------
  subroutine horizontalSlice(poly, yLo, yHi, subPolys&
  , nSubs, nPoints, status)
  !---------------------------------------------------------------------
  ! NOTE that it is ABSOLUTELY REQUIRED in this subroutine that the points in the poly are numbered in clockwise order and that none of the sides cross each other.
  !
  ! The sub has been copied from verticalSlice with minimal alterations. Therefore if bugs are found in one, the other should also be altered.

!    use mathutils_mod

    implicit none

    type(localPolyType), intent(in) :: poly(:)
    real, intent(in) :: yLo, yHi
!    type(localPolyType), intent(out) :: subPolys(size(poly),2*size(poly))
type(localPolyType), intent(out) :: subPolys(:,:)
! The dimensions given are the maximum possible limits. The first index is the number of the sub-poly resulting from the intersection of poly with the two verticals, the second index is the number of the point within that sub-poly.
!    integer, intent(out) :: nPoints(size(poly)), nSubs, status
integer, intent(out) :: nPoints(:), nSubs, status

    type :: intersectionType
      real :: x
      logical :: exists, isAvailable
    end type intersectionType
    type(intersectionType) :: interLo(size(poly)), interHi(size(poly))
    integer :: i, j, k, zone(size(poly)), nVertices, precedingVertex, state, indexVec(1)
    integer, parameter :: LOW = 1, HIGH = 2, MIDDLE = 3
    real :: yFrac
    logical :: vertexIsAvailable(size(poly)), mask(size(poly))
logical :: flag = .false.

    !-------------------------------------------------------------------
    ! Start

if (flag) then
print *, '      -----------------------------------------------------------'
print *, '      Now inside horizontalSlice.'
print *, ''
end if

    status = 0
    nPoints = 0 ! vector default
    nSubs = 0 ! default

if (flag) then
print *, '      minval(poly%pos%y):', minval(poly%pos%y)
print *, '      maxval(poly%pos%y):', maxval(poly%pos%y)
print *, '                     yLo:', yLo
print *, '                     yHi:', yHi
end if

  if (minval(poly%pos%y) .ge. yHi .or. maxval(poly%pos%y) .le. yLo) return

    nVertices = size(poly)

if (flag) then
print *, '               nVertices:', nVertices
print *, ''
end if

! Find out into which zone each poly vertex falls:

    do i = 1, nVertices
      if (poly(i)%pos%y .le. yLo) then
        zone(i) = LOW
      else if (poly(i)%pos%y .ge. yHi) then
        zone(i) = HIGH
      else
        zone(i) = MIDDLE
      end if
    end do

! Find which sides of the polygon intersect the left and right verticals:

    interLo%isAvailable = .false.
    interHi%isAvailable = .false.
    interLo%exists = .false.
    interHi%exists = .false.
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices) ! ie j=1 for i=nVertices, j=i+1 else.

      if ( (poly(i)%pos%y .le. yLo .and. poly(j)%pos%y .gt. yLo)&
      .or. (poly(i)%pos%y .gt. yLo .and. poly(j)%pos%y .le. yLo)) then
        yFrac = (yLo - poly(i)%pos%y) / (poly(j)%pos%y - poly(i)%pos%y)
        interLo(i)%x = fracVal(poly(i)%pos%x, poly(j)%pos%x, yFrac)
        interLo(i)%exists      = .true.
        interLo(i)%isAvailable = .true.
      end if
      if ( (poly(i)%pos%y .ge. yHi .and. poly(j)%pos%y .lt. yHi)&
      .or. (poly(i)%pos%y .lt. yHi .and. poly(j)%pos%y .ge. yHi)) then
        yFrac = (yHi - poly(i)%pos%y) / (poly(j)%pos%y - poly(i)%pos%y)
        interHi(i)%x = fracVal(poly(i)%pos%x, poly(j)%pos%x, yFrac)
        interHi(i)%exists      = .true.
        interHi(i)%isAvailable = .true.
      end if
    end do

    ! And, just for neatness really, change a glancing touch onto one of the horizontals from a non-MIDDLE point to a MIDDLE point. If this is not done, the intersection polygon will have 2 points at that place.
    k = 1 + modulo(1, nVertices)
    do i = 1, nVertices
      j = k				! ie j = i + 1, modulo nVertices
      k = 1 + modulo(i + 1, nVertices)	! ie k = i + 2, modulo nVertices
      if (zone(i) .eq. MIDDLE .and. zone(k) .eq. MIDDLE) then
        if (poly(j)%pos%y .eq. yLo) then
          if (isConvex(poly(i)%pos, poly(j)%pos, poly(k)%pos)) then
            zone(j) = MIDDLE
            interLo(i)%exists = .false.
            interLo(j)%exists = .false.
            interLo(i)%isAvailable = .false.
            interLo(j)%isAvailable = .false.
          end if
        else if (poly(j)%pos%y .eq. yHi) then
          if (isConvex(poly(i)%pos, poly(j)%pos, poly(k)%pos)) then
            zone(j) = MIDDLE
            interHi(i)%exists = .false.
            interHi(j)%exists = .false.
            interHi(i)%isAvailable = .false.
            interHi(j)%isAvailable = .false.
          end if
        end if
      end if
    end do

if (any(interLo%exists .and. interHi%exists .and. (zone .eq. MIDDLE))) then
!***** BUG!!
status = -1
return
end if

    vertexIsAvailable = .false. ! vector default
    where(zone .eq. MIDDLE) vertexIsAvailable = .true.

    ! Finally, because it is possible for poly()pos%y to be just on the other side of a horizontal to its neighbouring vertices, thereby generating two neighbouring intercepts with that horizontal, but these intercepts can nevertheless due to numerical rounding errors be at the same x value, the following is done to erase any such paired, same-x intercepts:
    do i = 1, nVertices
      j = 1 + modulo(i, nVertices)	! ie j = i + 1, modulo nVertices

      if (interLo(i)%exists .and. interLo(j)%exists&
      .and. interLo(i)%x .eq. interLo(j)%x) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interLo(i)%exists = .false.
        interLo(j)%exists = .false.
        interLo(i)%isAvailable = .false.
        interLo(j)%isAvailable = .false.
if (flag) print *, 'taken out LO intercepts', i, 'and', j
      end if

      if (interHi(i)%exists .and. interHi(j)%exists&
      .and. interHi(i)%x .eq. interHi(j)%x) then
        vertexIsAvailable(j) = .not. vertexIsAvailable(j)
        interHi(i)%exists = .false.
        interHi(j)%exists = .false.
        interHi(i)%isAvailable = .false.
        interHi(j)%isAvailable = .false.
if (flag) print *, 'taken out HI intercepts', i, 'and', j
      end if
    end do

if (flag) then
do i = 1, nVertices
  print *, '                            i:', i
  print *, '                poly(i)%pos%x:', poly(i)%pos%x
  print *, '                poly(i)%pos%y:', poly(i)%pos%y
  print *, '                      zone(i):', zone(i)
  print *, '         vertexIsAvailable(i):', vertexIsAvailable(i)
  print *, '            interLo(i)%exists:', interLo(i)%exists
  print *, '       interLo(i)%isAvailable:', interLo(i)%isAvailable
  print *, '                 interLo(i)%x:', interLo(i)%x
  print *, '            interHi(i)%exists:', interHi(i)%exists
  print *, '       interHi(i)%isAvailable:', interHi(i)%isAvailable
  print *, '                 interHi(i)%x:', interHi(i)%x
  print *, ''
end do

print *, '      Ok, now constructing the intersection:'
print *, ''
end if

! Now construct the intersection of the horizontals and the poly (could be a number of subpolys):

    do while(any(interLo%isAvailable) .or. any(interHi%isAvailable)&
    .or. any(vertexIsAvailable))

      ! Start a new sub-poly:

      nSubs = nSubs + 1
      nPoints(nSubs) = 1

if (flag) then
print *, '      Start ',nSubs,'nth new sub-poly:'
print *, ''
end if
      if (any(interLo%isAvailable)) then ! find it:
        do i = 1, nVertices
          if (interLo(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interLo(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%pos%y = yLo
        subPolys(nSubs, nPoints(nSubs))%pos%x = interLo(precedingVertex)%x

        if (zone(precedingVertex) .eq. LOW) then
          state = MIDDLE
        else
          state = LOW
        end if

if (flag) then
print *, '      Ok, I found some available lo-y intercepts.'
print *, ''
print *, '      precedingVertex for the first is: ', precedingVertex
print *, ''
end if
      else if (any(interHi%isAvailable)) then
        do i = 1, nVertices
          if (interHi(i)%isAvailable) then
            precedingVertex = i
        exit
          end if
        end do
        interHi(precedingVertex)%isAvailable = .false.
        subPolys(nSubs, nPoints(nSubs))%pos%y = yHi
        subPolys(nSubs, nPoints(nSubs))%pos%x = interHi(precedingVertex)%x

        if (zone(precedingVertex) .eq. HIGH) then
          state = MIDDLE
        else
          state = HIGH
        end if

if (flag) then
print *, '      Didnt find any available lo-y intercepts, but I found some hi-y intercepts available.'
print *, ''
print *, '      precedingVertex for the first is: ', precedingVertex
print *, ''
end if
      else ! there must be an available vertex.
        do i = 1, nVertices
          if (vertexIsAvailable(i)) then
            precedingVertex = i
        exit
          end if
        end do
        vertexIsAvailable(precedingVertex) = .false.
        subPolys(nSubs, nPoints(nSubs))%pos = poly(precedingVertex)%pos

        state = MIDDLE
if (flag) then
print *, '      Didnt find any available lo-y or hi-y intercepts, but I found some available polygon vertices.'
print *, ''
print *, '      Vertex of the first is: ', precedingVertex
print *, ''
end if
      end if

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'zone(precedingVertex=',precedingVertex,'): ', zone(precedingVertex)
print *, 'state: ', state
print *, ''
end if

      ! Now follow it around until the end:
      do
        select case(state)
        case(LOW)
if (flag) print *, 'state is LOW (=',LOW,'). Check next lower-x intercept on LO.'
        ! Move down until next lower intercept, after which state ought to
        ! be MIDDLE.
! NOTE!! This is the REVERSE to verticalSlice.
          mask = (interLo%exists&
          .and. interLo%x .lt. subPolys(nSubs, nPoints(nSubs))%pos%x)
!          .and. interLo%x .le. subPolys(nSubs, nPoints(nSubs))%pos%x)
if (.not. any(mask)) then
!***** BUG!!
status = -2
return
end if

!do i = 1, nVertices
!print *, '', i, interLo(i)%exists, mask(i)
!end do

          indexVec = maxloc(interLo%x, mask)
          precedingVertex = indexVec(1)
if (flag) print *, 'precedingVertex is now: ', precedingVertex
!!!      if (.not. interLo(precedingVertex)%isAvailable) exit ! because this
!!!      ! subpoly is now finished.
if (.not. interLo(precedingVertex)%isAvailable) then
if (flag) print *, 'Next lower-x intercept on LO is not available, poly must be finished.'
exit
end if

if (zone(precedingVertex) .ne. LOW) then
!***** BUG!!
status = -3
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interLo(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%pos%y = yLo
          subPolys(nSubs, nPoints(nSubs))%pos%x = interLo(precedingVertex)%x

          state = MIDDLE

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

        case(HIGH)
if (flag) print *, 'state is HIGH (=',HIGH,'). Check next higher-x intercept on HI.'
        ! Move up until next higher intercept, after which state ought to
        ! be MIDDLE.
! NOTE!! This is the REVERSE to verticalSlice.
          mask = (interHi%exists&
          .and. interHi%x .gt. subPolys(nSubs, nPoints(nSubs))%pos%x)
!          .and. interHi%x .ge. subPolys(nSubs, nPoints(nSubs))%pos%x)
if (.not. any(mask)) then
!***** BUG!!
status = -4
return
end if
          indexVec = minloc(interHi%x, mask)
          precedingVertex = indexVec(1)
if (flag) print *, 'precedingVertex is now: ', precedingVertex
!!!      if (.not. interHi(precedingVertex)%isAvailable) exit ! because this
!!!      ! subpoly is now finished.
if (.not. interHi(precedingVertex)%isAvailable) then
if (flag) print *, 'Next higher-x intercept on HI is not available, poly must be finished.'
exit
end if

if (zone(precedingVertex) .ne. HIGH) then
!***** BUG!!
status = -5
return
end if

          nPoints(nSubs) = nPoints(nSubs) + 1
          interHi(precedingVertex)%isAvailable = .false.
          subPolys(nSubs, nPoints(nSubs))%pos%y = yHi
          subPolys(nSubs, nPoints(nSubs))%pos%x = interHi(precedingVertex)%x

          state = MIDDLE

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

        case(MIDDLE)
if (flag) print *, 'state is MIDDLE (=',MIDDLE,').'
          if (interLo(precedingVertex)%isAvailable) then
if (flag) print *, 'this side intersects yLo.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            interLo(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%pos%y = yLo
            subPolys(nSubs, nPoints(nSubs))%pos%x = interLo(precedingVertex)%x

            state = LOW

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: LOW'
print *, ''
end if

          else&
          if (interHi(precedingVertex)%isAvailable) then
if (flag) print *, 'this side intersects yHi.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            interHi(precedingVertex)%isAvailable = .false.
            subPolys(nSubs, nPoints(nSubs))%pos%y = yHi
            subPolys(nSubs, nPoints(nSubs))%pos%x = interHi(precedingVertex)%x

            state = HIGH

if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: HIGH'
print *, ''
end if

          else
            precedingVertex = 1 + modulo(precedingVertex, nVertices)
if (flag) print *, 'vertex is now: ', precedingVertex
!!!      if (.not. vertexIsAvailable(precedingVertex)) exit ! because this
!!!      ! subpoly is now finished.
if (.not. vertexIsAvailable(precedingVertex)) then
if (flag) print *, 'Next vertex is not available, poly must be finished.'
exit
end if

if (flag) print *, 'this side doesnt intersect either yLo or yHi.'
            nPoints(nSubs) = nPoints(nSubs) + 1
            vertexIsAvailable(precedingVertex) = .false.
            subPolys(nSubs, nPoints(nSubs))%pos = poly(precedingVertex)%pos

!           state = MIDDLE ! - still.
if (flag) then
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%x: ', subPolys(nSubs, nPoints(nSubs))%pos%x
print *, 'subPolys(',nSubs,',',nPoints(nSubs),')%pos%y: ', subPolys(nSubs, nPoints(nSubs))%pos%y
print *, 'state: MIDDLE'
print *, ''
end if

          end if

        case default
!***** BUG!!
          status = -6
  return
        end select
if (flag) print *, ''
      end do ! complete sub-poly
if (flag) print *, ''
    end do ! no more available vertices or intersections.
if (flag) print *, ''

  end subroutine horizontalSlice

  !---------------------------------------------------------------------
  function areaOfPoly(poly) result(area)
  !---------------------------------------------------------------------
  ! Surprise surprise, calculates the area of the given polygon. The basic
  ! formula is derived (not by me, I got it off the comp.graphics.algorithms
  ! FAQ, but it is pretty obvious) by considering the polygon as two curves
  ! on the x-y plane. Each curve is a sequence of line segments and the
  ! curves are joined at both low-x and hi-x extrema. This is a definition
  ! for a polygon, right?? Now the area of the polygon can be easily seen
  ! to be equal to the integral of the lower curve subtracted from that of
  ! the upper curve. But the integral under a line segment from (x0,y0) to
  ! (x1,y1) is just
  !
  !	A = 0.5 (x1 - x0) * (y1 + y0).
  !
  ! Note that this is negative if x1 < x0. The total area of the polygon Ap
  ! can thus be expressed as the sum of the integrals of all its line
  ! segments: the 'bottom' line segments contribute the necessary negative
  ! factor. Thus
  !
  !	Ap = 0.5 sum_{i=1}^{N} ((x_i - x_{i-1}) * (y_i + y_{i-1}))
  !
  ! where 'i-1' = N if i = 1. This can be boiled down to
  !
  !	Ap = 0.5 sum_{i=1}^{N} (x_i * (y_{i-1} - y_{i+1}))
  !
  ! which involves fewer operations. Note that the answer is negative if
  ! the points in the polynomial are numbered in an anticlockwise sense.

    implicit none

    type(pointType), intent(in) :: poly(:)
    real :: area

    integer :: i, numPoints

    !-------------------------------------------------------------------
    ! Start

    area = 0.0
    numPoints = size(poly)
  if (numPoints .le. 2) return

    area = area + poly(1)%x * (poly(numPoints)%y - poly(2)%y)
    do i = 2, numPoints - 1
      area = area + poly(i)%x * (poly(i - 1)%y - poly(i + 1)%y)
    end do
    area = area + poly(numPoints)%x * (poly(numPoints - 1)%y - poly(1)%y)
    area = area / 2.0

  end function areaOfPoly

  !---------------------------------------------------------------------
  logical function pointIsInsidePoly(x, y, poly, includeBorder)
  !---------------------------------------------------------------------

!    use mathutils_mod

    implicit none

    real, intent(in) :: x, y
    type(localPolyType), intent(in) :: poly(:)
    logical, intent(in), optional :: includeBorder

    logical :: excludeBorder
    real :: intersectionY(size(poly)), xFrac
    integer :: numIntersections, i, numHigherIntersections

    !-------------------------------------------------------------------
    ! Start

    if (present(includeBorder)) then
      excludeBorder = .not. includeBorder
    else
      excludeBorder = .true. ! ie point is stated to be inside the poly only
      ! if it is neither outside nor on the border. In other words, the
      ! border is considered to be outside the poly.
    end if

    ! Calculate the y values of all points of intersection between polygon
    ! sides and a vertical line passing through x:
    numIntersections = 0
    do i = 1, size(poly) - 1
      if (poly(i)%pos%x .eq. x .and. poly(i)%pos%y .eq. y) then
        if (excludeBorder) then
          pointIsInsidePoly = .false.
        else
          pointIsInsidePoly = .true.
        end if
  return
      end if

      if (poly(i)%pos%x .le. x .and. poly(i + 1)%pos%x .gt. x) then
        numIntersections = numIntersections + 1
        xFrac = (x - poly(i)%pos%x) / (poly(i + 1)%pos%x - poly(i)%pos%x)
        intersectionY(numIntersections)&
        = fracVal(poly(i)%pos%y, poly(i + 1)%pos%y, xFrac)
      end if
    end do

    if (numIntersections .le. 0) then
      pointIsInsidePoly = .false.
  return
    end if

    if (any(intersectionY(1:numIntersections) .eq. y)) then
      if (excludeBorder) then
        pointIsInsidePoly = .false.
      else
        pointIsInsidePoly = .true.
      end if
  return
    end if

    ! Now find the number of intersections that have y value greater than y.
    ! If this is even, then (x,y) lies outside the poly; if odd, (x,y) lies
    ! within.
    numHigherIntersections = count(intersectionY(1:numIntersections) .gt. y)
    if (isOdd(numHigherIntersections)) then
      pointIsInsidePoly = .true.
    else
      pointIsInsidePoly = .false.
    end if

  end function pointIsInsidePoly

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module polygon
