!
! Module: samplepsf_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Makes an image of the (relative) source counts contained in an nxn sliding box.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module samplepsf_mod

  use dal
  use cal
  use param  
  use errorHandling
  use dss

  use constants
  use dal_aux
  use save_image
  use array_utils
  use ssc_misc
  use coordinate
!use coordinate_old
  use cal_aux
  use wcs_aux
  use polygon
  use dss_aux
  use geometric_types
  use splines
  use exposure
use dump	! for dumpImageToFits

  implicit none 

  private 

  ! Command-line parameters:
  real(double) ::&
    energyFraction
  character(255) ::&
    expImageSetName,&
    dssImageSetName,&
    indexImageSetName,&
    psfFracMapSetName,&
    tempImageName
  integer(int16) ::&
    boxHalfSize,& ! pixels in excess of the centre pixel. Eg for a 5x5 box,
    ! boxHalfSize should = 2. It is assumed that the box is square.
    numAxialSamples,&
    numRadialSamples

  ! Other globals:
  type(WcsT) ::&
    wcs
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp
  real(double) ::&
    mmToDetUnits, focalLengthMm

  public samplepsf

contains 

  !---------------------------------------------------------------------
  subroutine samplepsf
  !---------------------------------------------------------------------
  ! This convolves the exposure map with the PSF truncated in an nxn box
  ! to give a map of the relative source counts to be expected in an nxn
  ! sliding box.

    implicit none

    real(single), pointer :: expMap(:,:), convCubeData(:,:,:)
    integer(int32) ::&
      xSize, ySize, boxSize, numConvolvers, radialBin, axialBin, i
    real(double) :: psfEeV
    integer(int32), allocatable :: indexImage(:,:)
    real(single), allocatable :: convolvers(:,:,:,:), psfFracMap(:,:)
    type(DataSetT) :: dssImageSet, convSet
    type(DssT) :: dssPointer
    character(8) :: wcsType
    type(ArrayT) :: indexArray, convCube
    integer(int32), pointer :: arrayDataInt32(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call getCalInfo(expImageSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in cal_aux. Sets up cal for inst.

    call getWcsInfo(expImageSetName, 0, wcs)
!call dumpWcs(wcs)
!call error('','')
    call easifyWcsType(wcs%x%ctype, wcs%y%ctype, wcsType)
    if (wcsType .ne. 'DETXY' .and. wcsType .ne. 'XY')&
call error('badWcsCtype','')

    call readArrayData(expImageSetName, expMap) ! in fits_utils
    xSize = size(expMap, 1)
    ySize = size(expMap, 2)
    deallocate(expMap)

    dssImageSet = dataSet(dssImageSetName, READ)
    dssPointer = dataSubspace(block(dssImageSet, 0))

    call getPsfEeV(dssPointer, psfEeV)
    call release(dssImageSet)

    allocate(indexImage(xSize, ySize), psfFracMap(xSize, ySize))
    boxSize = 1 + 2 * boxHalfSize
    allocate(convolvers(boxSize, boxSize, numRadialSamples, numAxialSamples))

    call makeConvolvers(psfEeV, wcs, wcsType&
      , instrumentId, expStartTimeStamp, scAttitude&
      , energyFraction, indexImage, convolvers, psfFracMap)

    call message(AppMsg, verbose, 'Saving the output image.')

    call saveImage(psfFracMap, expImageSetName, psfFracMapSetName, tempImageName)
    ! in save_image

    call message(AppMsg, verbose, 'Saving the PSF convolvers.')

    ! Save convolvers and index image together. Note that this file format
    ! is designed to be understood by asmooth.

    convSet = dataSet(indexImageSetName, CREATE)

    indexArray = addArray(convSet, 'INDEXIMG', INTEGER32, (/xSize, ySize/))
    arrayDataInt32 => int32Array2Data(indexArray)
    arrayDataInt32 = indexImage
    call release(indexArray)
    deallocate(indexImage)

    numConvolvers = numRadialSamples * numAxialSamples
    convCube = addArray(convSet, 'CONV_000', REAL32&
      , (/boxSize, boxSize, numConvolvers/))
    convCubeData => real32Array3Data(convCube)
    do radialBin = 1, numRadialSamples
      do axialBin = 1, numAxialSamples
        i = ((radialBin - 1) * numAxialSamples) + axialBin
        convCubeData(:,:,i-1) = convolvers(:,:,radialBin,axialBin)&
          / sum(convolvers(:,:,radialBin,axialBin))
        ! Note that the convolvers are normalized.
      end do
    end do

    call release(convSet)

  end subroutine samplepsf

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    boxHalfSize      = intParameter('boxhalfsize')
    numAxialSamples  = intParameter('naxial')
    numRadialSamples = intParameter('nradial')
    expImageSetName   = stringParameter('expimageset')
    indexImageSetName = stringParameter('convolversset')
    dssImageSetName   = stringParameter('dssimageset')
    psfFracMapSetName = stringParameter('psffracmapset')
    energyFraction = realParameter('efraction')
    tempImageName   = stringParameter('tempset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getPsfEeV(dssPointer, psfEeV)
  !---------------------------------------------------------------------
  ! Calculate the energy at which to obtain the PSF by taking the mean of
  ! the energy selection as recorded in the image dss.
  !
  ! NOTE! The cal needs to be set up for the correct instrument before
  ! calling this subroutine.

    implicit none

    type(DssT), intent(in) :: dssPointer
    real(double), intent(out) :: psfEeV

    type(RangeT), pointer :: piRanges(:), EeVranges(:)
!    real(single), pointer :: ePointer(:)
    real(single) :: EeVsamples(100)
!integer :: i
    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, verbose, 'Calculating the mean of the selected &
      &energy band from the selection ranges stored in the DSS.')

    call readRanges(dssPointer, 'PI', piRanges)
if (size(piRanges) <= 0) call error('','')
    call correctRanges(piRanges)
    EeVranges => convertPiRangesToEeV(piRanges) ! the call deallocates piRanges.
    call sampleEnergyRange(EeVranges, EeVsamples)

    psfEeV = integrateRanges(EeVsamples, EeVsamples, EeVranges)&
      / sum(EeVranges%upper%value - EeVranges%lower%value)
    ! The denominator looks funny, but if you analyse the whole expression
    ! you find it actually evaluates to (a numerical approximation of)
    !
    !	  _N     EeVranges(i)%upper%value
    !	 \          /
    !	  >         | E dE
    !	 /__        /
    !	  i=1    EeVranges(i)%lower%value
    !	----------------------------------
    !	  _N     EeVranges(i)%upper%value
    !	 \          /
    !	  >         | dE
    !	 /__        /
    !	  i=1    EeVranges(i)%lower%value
    !
    ! In other words, this is the average value of EeV in the set of ranges.

    deallocate(EeVranges)

  end subroutine getPsfEeV

  !---------------------------------------------------------------------
  subroutine makeConvolvers(psfEeV, wcs, wcsType&
  , instrumentId, timeStamp, scAttitude&
  , energyFraction, indexImage, convolvers, psfFracMap)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up for the correct instrument before this
  ! is called.

    implicit none

    real(double),                 intent(in)  :: psfEeV
    type(WcsT),                   intent(in)  :: wcs
!type(WcsT),            intent(inout)  :: wcs
    character(*),                 intent(in)  :: wcsType
    integer(int32),               intent(in)  :: instrumentId
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp,&
                                                 energyFraction
    integer(int32),               intent(out) :: indexImage(:,:)
    real(single),                 intent(out) :: convolvers(:,:,:,:),&
                                                 psfFracMap(size(indexImage,1)&
                                                          , size(indexImage,2))

    integer(int16) ::&
      xSize, ySize, boxXsize, boxYsize,&
      numRadialBins, numAxialBins, radialBin, axialBin, i
    integer(int16), pointer :: indices(:)
    real(double), allocatable :: thetaMap(:,:), phiMap(:,:)
    integer(int32) :: oldAccuracyLevel
    type(PsfValidityRangesT) :: validPsfRanges
    real(double) ::&
      thetaArcsecLo, psfThetaArcsec, phiLo, psfPhi
    type(PsfT) :: psfPointer
    type(PsfBinSizeT) :: psfPixelSizeMm
    type(PsfWindowT) :: psfWindow
    real(single), pointer :: psfImage(:,:)
    type(Point2dT), pointer :: psfImagePixelCorners(:,:)
    type(Point2dT) :: psfCentrePixels
      ! type defined in coordinate.f90.
    integer :: foreignStatus
    real(single) ::&
      boxCentreX, boxCentreY,&
      psfFracSamplesArray(size(convolvers, 3), size(convolvers, 4)+1),&
      thetaSamplesVector(size(convolvers, 3)),&
      phiSamplesVector(size(convolvers, 4)+1),&
      thetaBinLoEdges(size(convolvers, 3)),&
      phiBinLoEdges(size(convolvers, 4)+1)
    real(double), parameter :: maxThetaArcsec = 900.0
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, verbose, 'Converting image coordinates to TELCOORD.')

    xSize = size(indexImage, 1)
    ySize = size(indexImage, 2)
    boxXsize = size(convolvers, 1)
    boxYsize = size(convolvers, 2)
    boxCentreX = (boxXsize + 1) / 2.0
    boxCentreY = (boxYsize + 1) / 2.0
    numRadialBins = size(convolvers, 3)
    numAxialBins = size(convolvers, 4)

    allocate(thetaMap(xSize, ySize), phiMap(xSize, ySize))
    call getThetaPhiMaps(wcs, scAttitude, timeStamp, thetaMap, phiMap)

    call CAL_getState(accuracyLevel=oldAccuracyLevel)
    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
    mmToDetUnits = 3600.0 / focalLengthMm / PiOn180 / detUnitArcsec
    ! Note that detUnitArcsec is defined in coordinate.f90.

    validPsfRanges = CAL_psfValidityRanges()

    call message(AppMsg, verbose, 'Creating PSF convolution kernels.')

!    indices => indicize(0_int16, numRadialBins-1_int16)
allocate(indices(numRadialBins))
indices = (/(i, i=0,numRadialBins-1)/)
    thetaSamplesVector = fracVal(0.0, real(maxval(thetaMap))&
      , indices / real(numRadialBins-1))
    thetaBinLoEdges = fracVal(0.0, real(maxval(thetaMap))&
      , (indices - 0.5) / real(numRadialBins-1))
    deallocate(indices)

!    indices => indicize(0_int16, numAxialBins)
allocate(indices(numAxialBins))
indices = (/(i, i=0,numAxialBins-1)/)
    phiSamplesVector = fracVal(0.0, real(TwoPi), indices / real(numAxialBins))
    phiBinLoEdges = fracVal(0.0, real(TwoPi)&
      , (indices - 0.5) / real(numAxialBins))
    deallocate(indices)

    do radialBin = 1, numRadialBins
print *, 'radialBin/numRadialBins:', radialBin,'/',numRadialBins

      thetaArcsecLo = thetaBinLoEdges(radialBin)
      psfThetaArcsec = thetaSamplesVector(radialBin)

      if (psfThetaArcsec > validPsfRanges%thetaMax) then
        psfThetaArcsec = validPsfRanges%thetaMax&
          * (1.0d0 - epsilon(psfThetaArcsec))
      end if

      do axialBin = 1, numAxialBins
        i = ((radialBin - 1) * numAxialBins) + axialBin
        phiLo = phiBinLoEdges(axialBin)
        where(thetaMap >= thetaArcsecLo .and. phiMap >= phiLo) indexImage = i

        psfPhi = phiSamplesVector(axialBin)

        psfPointer = CAL_getPSF(psfEeV, psfThetaArcsec, psfPhi)
        psfPixelSizeMm = CAL_psfBinSize(psfPointer)
        psfWindow = CAL_psfEnboxedEnergy(psfPointer, fraction=energyFraction)

        call CAL_psfImage(psfPointer, psfWindow, psfImage)

        ! Normalize:
        psfImage = psfImage / (sum(psfImage) / energyFraction)

        call getPsfImagePixelCorners(psfThetaArcsec, psfPhi&
          , psfPixelSizeMm, psfImage, wcs, wcsType&
          , instrumentId, timeStamp, scAttitude&
          , psfImagePixelCorners, psfCentrePixels) ! The pairs of numbers
        ! (type(Point2dT)) in this array represent the coordinates of the
        ! corners of the psfImage pixels, expressed in the exposure-map
        ! image (ie, sky) coordinate system.

        ! Shift the array of PSF-image pixel corners by integer sky-pixel
        ! widths until the PSF centre is as near as possible to the centre
        ! of convolvers():

        psfImagePixelCorners%x = psfImagePixelCorners%x&
          + boxCentreX - psfCentrePixels%x
        psfImagePixelCorners%y = psfImagePixelCorners%y&
          + boxCentreY - psfCentrePixels%y

        call regrid(psfImagePixelCorners, psfImage&
          , convolvers(:,:,radialBin,axialBin), status=foreignStatus)
        ! This rebins the psfImage values from its original CAMCOORD2 system
        ! to the exposure-map image (ie, sky) coordinate system. The array
        ! returned in convolvers() is centred as near as possible on the
        ! centre of the PSF.

if (foreignStatus < 0) call error('','')
!        if (foreignStatus > 0) then
!          print *, 'regrid exit status:', foreignStatus
!        end if

!**** is there a status flag to indicate if the psf didn't cover all the convolvers array? If so, protest and advise use of a larger energyFraction.

        call CAL_releaseMemory(psfImage)

        psfFracSamplesArray(radialBin,axialBin)&
          = sum(convolvers(:,:,radialBin,axialBin))
      end do

      i = ((radialBin - 1) * numAxialBins) + 1 ! ie, same as for 1st axialBin
      ! this radialBin. This splices up the indexImage over the phi=0/phi=2pi
      ! join.

      phiLo = phiBinLoEdges(numAxialBins+1)
      where(thetaMap >= thetaArcsecLo .and. phiMap >= phiLo) indexImage = i

      psfFracSamplesArray(radialBin, numAxialBins+1)&
        = psfFracSamplesArray(radialBin, 1)
    end do

    call CAL_setState(accuracyLevel=oldAccuracyLevel)
    deallocate(psfImagePixelCorners)
!call dumpImageToFits(indexImage, 'indexImage.ds')
!call error('testStop','')

call interpolateInTelcoord(thetaSamplesVector, phiSamplesVector&
  , psfFracSamplesArray, thetaMap, phiMap, psfFracMap)

    call CAL_setState(accuracyLevel=oldAccuracyLevel)

  end subroutine makeConvolvers

  !---------------------------------------------------------------------
  subroutine interpolateInTelcoord(thetaSamplesVector, phiSamplesVector&
  , psfFracSamplesArray, thetaMap, phiMap, psfFracMap)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      thetaSamplesVector(:), phiSamplesVector(:),&
      psfFracSamplesArray(size(thetaSamplesVector),size(phiSamplesVector))
    real(double), intent(in) ::&
      thetaMap(:,:), phiMap(size(thetaMap, 1), size(thetaMap, 2))
    real(single), intent(out) ::&
      psfFracMap(size(thetaMap, 1), size(thetaMap, 2))

    integer(int16) ::&
      numThetaSamples, numPhiSamples,&
      xsize, ysize, loPhiI, hiPhiI,&
      xi, yi, phiI, i, j
    real(single), allocatable ::&
      secondDerivWrtTheta(:,:)
    real(single) :: thetaArcsec, phi, loPhiValue, hiPhiValue, phiFrac

    !-------------------------------------------------------------------
    ! Start

    numThetaSamples = size(thetaSamplesVector)
    numPhiSamples   = size(phiSamplesVector)
    ! NOTE these all must be >= 2.

    xsize = size(psfFracMap, 1)
    ysize = size(psfFracMap, 2)

    if  (minval(thetaMap) < thetaSamplesVector(1)) then
print *, 'minval(thetaMap), thetaSamplesVector(1): ', minval(thetaMap), thetaSamplesVector(1)
call error('bug','')
    end if

    if  (minval(phiMap) < phiSamplesVector(1)&
    .or. maxval(phiMap) > phiSamplesVector(numPhiSamples)) then
print *, 'minval(phiMap), phiSamplesVector(1): ', minval(phiMap), phiSamplesVector(1)
print *, 'maxval(phiMap), phiSamplesVector(n): ', maxval(phiMap), phiSamplesVector(numThetaSamples)
call error('bug','')
    end if

    loPhiI =  largestTrueIndex(logical(phiSamplesVector <= minval(phiMap),bool))
    hiPhiI = smallestTrueIndex(logical(phiSamplesVector >= maxval(phiMap),bool))

    if (loPhiI == hiPhiI) then ! unlikely as it implies that
      ! there is only 1 value of phi among the image pixels, and
      ! also that that value lies exactly on one of the samples
      ! in phiSamplesVector. Nevertheless, could happen I suppose.

      if (hiPhiI < numPhiSamples) then
        hiPhiI = hiPhiI + 1
      else
        loPhiI = loPhiI - 1
      end if
    end if

    ! Preload the 2nd derivatives wrt theta for the cubic spline:

    allocate(secondDerivWrtTheta(numThetaSamples, loPhiI:hiPhiI))
    do phiI = loPhiI, hiPhiI
      call calcCubicSplines(thetaSamplesVector, psfFracSamplesArray(:, phiI)&
        , loBoundType='fixed1st', loBoundValue=0.0&
        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
        , y2=secondDerivWrtTheta(:, phiI)&
      )
    end do

    ! Now interpolate the psf fraction array for all image pixels:

    psfFracMap = 0.0
    do xi = 1, xsize
      do yi = 1, ysize
        thetaArcsec = thetaMap(xi, yi)	! type conversion from
        phi = phiMap(xi, yi)		! double to single.
      if (thetaArcsec > thetaSamplesVector(numThetaSamples)) cycle

        ! Get the indices of the lower LH corner of the theta-phi
        ! rectangle in theta-phi space that contains both thetaArcsec
        ! and phi:

        i = largestTrueIndex(logical(thetaSamplesVector <= thetaArcsec, bool))
        if (i == numThetaSamples) i = numThetaSamples - 1
        j = largestTrueIndex(logical(phiSamplesVector <= phi, bool))
        if (j == hiPhiI) j = hiPhiI - 1

        ! Get the cubic-spline interpolation at each of the two
        ! bracketting phi values:

        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
          , psfFracSamplesArray(i:i+1, j)&
          , secondDerivWrtTheta(i:i+1, j),   thetaArcsec, loPhiValue)
        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
          , psfFracSamplesArray(i:i+1, j+1)&
          , secondDerivWrtTheta(i:i+1, j+1), thetaArcsec, hiPhiValue)

        ! Finally do a linear interpolation between these two phi values:

        phiFrac = (phi - phiSamplesVector(j))&
          / (phiSamplesVector(j+1) - phiSamplesVector(j))
        psfFracMap(xi, yi) = fracVal(loPhiValue, hiPhiValue, phiFrac)
      end do
    end do

    deallocate(secondDerivWrtTheta)

    if (any(thetaMap > thetaSamplesVector(numThetaSamples)))&
      call warning('thetaOutsideRange','')

  end subroutine interpolateInTelcoord

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module samplepsf_mod
