!
! Module: econvolverprep_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Makes an a series of convolvers, each of which is a portion of the PSF contained within a square nxn box of pixels. Each convolver is approximately correct for a given area of the input exposure map. An index image is also output to connect each convolver with its appropriate area.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module econvolverprep_mod

  use dal
  use cal
  use param  
  use errorHandling
  use dss

  use cal_aux		! for getCalInfo
  use constants
  use coordinate	! for getThetaPhiMaps, getPsfImagePixelCorners
  use dal_aux		! for readArrayData
  use dss_aux		! for readRanges
  use dss_ranges_aux	! for integrateRanges
  use exposure		! for convertPiToEeV, sampleEnergyRange
  use geometric_types	! for Point2dT
!  use polygon		! for regrid
  use regridders	! for regrid
  use ssc_misc		! for fracVal
  use wcs_aux		! for readWcs, easifyWcsType, writeWcs

!use dump	! for dumpImageToFits
!!use coordinate_old

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    expImageSetName = '',&
    dssImageSetName = '',&
    indexImageSetName = ''
!,&
!    psfFracMapSetName = ''
  character(20) ::&
    psfEnergyStyle = ''
  integer(int16) ::&
    boxHalfSize = 0,& ! pixels in excess of the centre pixel. Eg for a 5x5 box,
    ! boxHalfSize should = 2. It is assumed that the box is square.
    numAxialSamples = 0,&
    numRadialSamples = 0
  real(double) ::&
    userPsfEeV = 0.0d0,&
    energyFraction = 0.0d0 
!**** the task should calculate a value of energyFraction itself.

  ! Other globals:
  type(WcsT) ::&
    wcs
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp
  real(double) ::&
    mmToDetUnits
!, focalLengthMm

  public econvolverprep

contains 

  !---------------------------------------------------------------------
  subroutine econvolverprep
  !---------------------------------------------------------------------
  ! This prepares a set of convolvers which consist of the PSF appropriate
  ! to the position in the field of view, truncated by an nxn square box.
  ! These convolvers are intended to be supplied to asmooth in conjunction
  ! with an exposure map. Asmooth convolves the exposure map with the
  ! convolvers to give a map of the relative source counts to be expected
  ! in the nxn sliding box.
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: expMap(:,:), convCubeData(:,:,:)
    integer(int32) ::&
      xSize, ySize, boxSize, numConvolvers, radialBin, axialBin, i
    real(double) :: psfEeV
    integer(int32), allocatable :: indexImage(:,:)
    real(single), allocatable :: convolvers(:,:,:,:)
    type(DataSetT) :: dssImageSet, convSet
    type(DssT) :: dssPointer
    character(8) :: wcsType
    type(ArrayT) :: indexArray, convCube
    integer(int32), pointer :: arrayDataInt32(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call getCalInfo(expImageSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux. Sets up cal for inst.

    call readWcs(expImageSetName, wcs) ! in ssclib/wcs_aux.
    call easifyWcsType(wcs%x%ctype, wcs%y%ctype, wcsType)
    if (wcsType .ne. 'DETXY' .and. wcsType .ne. 'XY')&
call error('badWcsCtype',"")

    call readArrayData(expImageSetName, expMap) ! in ssclib/dal_aux
    xSize = size(expMap, 1)
    ySize = size(expMap, 2)
    deallocate(expMap)

    select case(psfEnergyStyle)
    case('user')
      psfEeV = userPsfEeV
    case('dss')
      dssImageSet = dataSet(dssImageSetName, READ)
      dssPointer = dataSubspace(block(dssImageSet, 0))
      call getPsfEeV(dssPointer, psfEeV)
      call release(dssImageSet)
    case default
call error('badPsfEnergyStyle',"The value of the parameter --psfestyle was &
&not recognized.")
    end select

    allocate(indexImage(xSize, ySize))
    boxSize = 1 + 2 * boxHalfSize
    allocate(convolvers(boxSize, boxSize, numRadialSamples, numAxialSamples))

    call makeConvolvers(psfEeV, wcs, wcsType&
      , instrumentId, expStartTimeStamp, scAttitude&
      , energyFraction, indexImage, convolvers)
!!!call makeConvolvers(psfEeV, wcs, expStartTimeStamp, scAttitude&
!!!  , energyFraction, indexImage, convolvers)

    call message(AppMsg, verbose, 'Saving the PSF convolvers.')

    ! Save convolvers and index image together. Note that this file format
    ! is designed to be understood by asmooth.

    convSet = dataSet(indexImageSetName, CREATE)

    indexArray = addArray(convSet, 'INDEXIMG', INTEGER32, (/xSize, ySize/))
    call writeWcs(indexArray, wcs) ! in ssclib/wcs_aux.
    arrayDataInt32 => int32Array2Data(indexArray)
    arrayDataInt32 = indexImage
    call release(indexArray)
    deallocate(indexImage)

    numConvolvers = numRadialSamples * numAxialSamples
    convCube = addArray(convSet, 'CONV_000', REAL32&
      , (/boxSize, boxSize, numConvolvers/))
    convCubeData => real32Array3Data(convCube)
    do radialBin = 1, numRadialSamples
      do axialBin = 1, numAxialSamples
        i = ((radialBin - 1) * numAxialSamples) + axialBin
        convCubeData(:,:,i-1) = convolvers(:,:,radialBin,axialBin)
      end do
    end do

    call release(convSet)

  end subroutine econvolverprep

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    boxHalfSize      = intParameter('boxhalfsize')
    numAxialSamples  = intParameter('naxial')
    numRadialSamples = intParameter('nradial')
    expImageSetName   = stringParameter('expmapset')
    indexImageSetName = stringParameter('convolversset')
    energyFraction = realParameter('efraction')

    psfEnergyStyle   = stringParameter('psfestyle')
    select case(psfEnergyStyle)
    case('user')
      userPsfEeV = realParameter('psfenergy')
    case('dss')
      dssImageSetName = stringParameter('dssimageset')
    case default
call error('badPsfEnergyStyle',"The value of the parameter --psfestyle was &
&not recognized.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getPsfEeV(dssPointer, psfEeV)
  !---------------------------------------------------------------------
  ! Calculate the energy at which to obtain the PSF by taking the mean of
  ! the energy selection as recorded in the image dss.
  !
  ! NOTE! The cal needs to be set up for the correct instrument before
  ! calling this subroutine.

    implicit none

    type(DssT), intent(in) :: dssPointer
    real(double), intent(out) :: psfEeV

    type(RangeT), pointer :: piRanges(:), EeVranges(:)
    real(double) :: EeVsamples(100)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, verbose, 'Calculating the mean of the selected &
      &energy band from the selection ranges stored in the DSS.')

    call readRanges(dssPointer, 'PI', piRanges) ! in ssclib/dss_aux

    if (size(piRanges) <= 0)&
call error('noDssPiRanges',"No filter on column PI was found in the &
&data subspace.")

    call correctRanges(piRanges) ! in ssclib/dss_aux
    EeVranges = convertPiToEeV(piRanges) ! in ssclib/exposure.
    deallocate(piRanges)

    call sampleEnergyRange(EeVranges, EeVsamples) ! in ssclib/exposure

    psfEeV = integrateRanges(real(EeVsamples, single)&
      , real(EeVsamples, single), EeVranges)&
      / sum(EeVranges%upper%value - EeVranges%lower%value)
    ! (integrateRanges is in ssclib/dss_ranges_aux.) The denominator looks
    ! funny, but if you analyse the whole expression you find it actually
    ! evaluates to (a numerical approximation of)
    !
    !	  _N     EeVranges(i)%upper%value
    !	 \          /
    !	  >         | E dE
    !	 /__        /
    !	  i=1    EeVranges(i)%lower%value
    !	----------------------------------
    !	  _N     EeVranges(i)%upper%value
    !	 \          /
    !	  >         | dE
    !	 /__        /
    !	  i=1    EeVranges(i)%lower%value
    !
    ! In other words, this is the average value of EeV in the set of ranges.

    deallocate(EeVranges)

  end subroutine getPsfEeV

  !---------------------------------------------------------------------
  subroutine makeConvolvers(psfEeV, wcs, wcsType&
  , instrumentId, timeStamp, scAttitude&
  , energyFraction, indexImage, convolvers)
!!!  subroutine makeConvolvers(psfEeV, wcs, timeStamp, scAttitude&
!!!  , energyFraction, indexImage, convolvers)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up for the correct instrument before this
  ! is called.

    implicit none

    real(double),                 intent(in)  :: psfEeV
    type(WcsT),                   intent(in)  :: wcs
character(*),                 intent(in)  :: wcsType
integer(int32),               intent(in)  :: instrumentId
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp,&
                                                 energyFraction
    integer(int32),               intent(out) :: indexImage(:,:)
    real(single),                 intent(out) :: convolvers(:,:,:,:)

    integer(int16) ::&
      xSize, ySize, boxXsize, boxYsize,&
      numRadialBins, numAxialBins, radialBin, axialBin, i
    integer(int16), pointer :: indices(:)
    real(double), allocatable :: thetaMap(:,:), phiMap(:,:)
    integer(int32) :: oldAccuracyLevel
    type(PsfValidityRangesT) :: validPsfRanges
    real(double) ::&
      thetaArcsecLo, psfThetaArcsec, phiLo, psfPhi
    type(PsfT) :: psfPointer
    type(PsfBinSizeT) :: psfPixelSizeMm
    type(PsfWindowT) :: psfWindow
    real(single), pointer :: psfImage(:,:)
    type(Point2dT), allocatable ::&
      psfImagePixelCorners(:,:)
!!!, psfImgPixCrnrDet(:,:)
!!!type(Point2dDbleT), allocatable ::&
!!!  psfImgPixCrnrTel(:,:), psfImgPixCrnrTan(:,:)
    type(Point2dT) :: psfCentrePixels
!!!, psfCentreDet
!!!type(Point2dDbleT) :: psfCentreTan
      ! types defined in coordinate.f90.
    integer :: foreignStatus
    real(single) ::&
      boxCentreX, boxCentreY,&
      thetaSamplesVector(size(convolvers, 3)),&
      phiSamplesVector(size(convolvers, 4)+1),&
      thetaBinLoEdges(size(convolvers, 3)),&
      phiBinLoEdges(size(convolvers, 4)+1)
    real(double), parameter :: maxThetaArcsec = 900.0
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, verbose, 'Converting image coordinates to TELCOORD.')

    xSize = size(indexImage, 1)
    ySize = size(indexImage, 2)
    boxXsize = size(convolvers, 1)
    boxYsize = size(convolvers, 2)
    boxCentreX = (boxXsize + 1) / 2.0
    boxCentreY = (boxYsize + 1) / 2.0
    numRadialBins = size(convolvers, 3)
    numAxialBins = size(convolvers, 4)

    allocate(thetaMap(xSize, ySize), phiMap(xSize, ySize))
    call getThetaPhiMaps(wcs, scAttitude, timeStamp, thetaMap, phiMap)
    ! in ssclib/coordinate

    call CAL_getState(accuracyLevel=oldAccuracyLevel)
    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)
!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
!    mmToDetUnits = 3600.0 / focalLengthMm / PiOn180 / detUnitArcsec
    ! Note that detUnitArcsec is defined in coordinate.f90.
mmToDetUnits = calcMmToDetConversionFactor() ! in ssclib/coordinate.f90.

    validPsfRanges = CAL_psfValidityRanges()

    call message(AppMsg, verbose, 'Creating PSF convolution kernels.')


    allocate(indices(numRadialBins))
    indices = (/(i, i=0,numRadialBins-1)/)
    thetaSamplesVector = fracVal(0.0, real(maxval(thetaMap))&
      , indices / real(numRadialBins-1)) ! in ssclib/ssc_misc
    thetaBinLoEdges = fracVal(0.0, real(maxval(thetaMap))&
      , (indices - 0.5) / real(numRadialBins-1)) ! in ssclib/ssc_misc
    deallocate(indices)

    allocate(indices(numAxialBins+1))
    indices = (/(i, i=0,numAxialBins)/)
    phiSamplesVector = fracVal(0.0, real(TwoPi), indices / real(numAxialBins))
    phiBinLoEdges = fracVal(0.0, real(TwoPi)&
      , (indices - 0.5) / real(numAxialBins)) ! in ssclib/ssc_misc
    deallocate(indices)

    do radialBin = 1, numRadialBins
print *, 'radialBin/numRadialBins:', radialBin,'/',numRadialBins
!*** make it a message

      thetaArcsecLo = thetaBinLoEdges(radialBin)
      psfThetaArcsec = thetaSamplesVector(radialBin)

      if (psfThetaArcsec > validPsfRanges%thetaMax) then
        psfThetaArcsec = validPsfRanges%thetaMax&
          * (1.0d0 - epsilon(psfThetaArcsec))
      end if

      do axialBin = 1, numAxialBins
        i = ((radialBin - 1) * numAxialBins) + axialBin
        phiLo = phiBinLoEdges(axialBin)
        where(thetaMap >= thetaArcsecLo .and. phiMap >= phiLo) indexImage = i
        psfPhi = phiSamplesVector(axialBin)

        psfPointer = CAL_getPSF(psfEeV, psfThetaArcsec, psfPhi)
        psfPixelSizeMm = CAL_psfBinSize(psfPointer)
        psfWindow = CAL_psfEnboxedEnergy(psfPointer, fraction=energyFraction)

        call CAL_psfImage(psfPointer, psfWindow, psfImage)

        ! Normalize:
        psfImage = psfImage / (sum(psfImage) / energyFraction)

        allocate(psfImagePixelCorners(size(psfImage,1)+1, size(psfImage,2)+1))
!!!allocate(psfImgPixCrnrTel(size(psfImage,1)+1, size(psfImage,2)+1))
!!!allocate(psfImgPixCrnrDet(size(psfImage,1)+1, size(psfImage,2)+1))
!!!allocate(psfImgPixCrnrTan(size(psfImage,1)+1, size(psfImage,2)+1))

        call getPsfImagePixelCorners_old(psfThetaArcsec, psfPhi&
          , psfPixelSizeMm, psfImage, wcs, wcsType&
          , instrumentId, timeStamp, scAttitude&
          , psfImagePixelCorners, psfCentrePixels) ! The pairs of numbers
        ! (type(Point2dT)) in this array represent the coordinates of the
        ! corners of the psfImage pixels, expressed in the exposure-map
        ! image (ie, sky) coordinate system. The routine is in
        ! ssclib/coordinate.

!!!call getPsfImagePixelCorners(psfThetaArcsec, psfPhi&
!!!  , psfPixelSizeMm, psfImage, psfImgPixCrnrTel)
!!!! in ssclib/coordinate.
!!!!
!!!! The pairs of numbers (type(Point2dT)) in this array represent the
!!!! coordinates of the corners of the psfImage pixels, expressed in
!!!! the exposure-map image (ie, sky) coordinate system.
!!!
!!!call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
!!!  , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
!!!! in ssclib/coordinate.
!!!
!!!call instToTan(psfImgPixCrnrTan%x, psfImgPixCrnrTan%y, wcs%x%refWorld&
!!!  , wcs%y%refWorld, scAttitude, timeStamp&
!!!  , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
!!!! in ssclib/coordinate.
!!!
!!!call tanToPixels(wcs, psfImgPixCrnrTan%x, psfImgPixCrnrTan%y&
!!!  , psfImagePixelCorners%x, psfImagePixelCorners%y)
!!!! in ssclib/wcs_aux.
!!!
!!!! Transform the position of the PSF centre:
!!!
!!!call telCoordToDetXY(psfThetaArcsec, psfPhi, psfCentreDet%x, psfCentreDet%y)
!!!! in ssclib/coordinate.
!!!
!!!call instToTan(psfCentreTan%x, psfCentreTan%y, wcs%x%refWorld&
!!!  , wcs%y%refWorld, scAttitude, timeStamp&
!!!  , psfCentreDet%x, psfCentreDet%y)
!!!! in ssclib/coordinate.
!!!
!!!call tanToPixels(wcs, psfCentreTan%x, psfCentreTan%y&
!!!  , psfCentrePixels%x, psfCentrePixels%y)
!!!! in ssclib/wcs_aux.


        ! Shift the array of PSF-image pixel corners by integer sky-pixel
        ! widths until the PSF centre is as near as possible to the centre
        ! of convolvers():

        psfImagePixelCorners%x = psfImagePixelCorners%x&
          + boxCentreX - psfCentrePixels%x
        psfImagePixelCorners%y = psfImagePixelCorners%y&
          + boxCentreY - psfCentrePixels%y

        call regridCartesian(psfImagePixelCorners, psfImage&
          , convolvers(:,:,radialBin,axialBin), status=foreignStatus)
        ! This rebins the psfImage values from its original CAMCOORD2 system
        ! to the exposure-map image (ie, sky) coordinate system. The array
        ! returned in convolvers() is centred as near as possible on the
        ! centre of the PSF. The routine is in ssclib/regridders.

        deallocate(psfImagePixelCorners)
!!!deallocate(psfImgPixCrnrTel)
!!!deallocate(psfImgPixCrnrDet)
!!!deallocate(psfImgPixCrnrTan)

if (foreignStatus < 0) call error('','')
!        if (foreignStatus > 0) then
!          print *, 'regrid exit status:', foreignStatus
!        end if

        if (any(convolvers(:,:,radialBin,axialBin) == 0.0))&
          call warning('psfImageTooSmall',"The PSF image doesn't cover &
          &all the convolver extent. You should use a larger value of &
          &--efraction.")
        ! This is not 100% adequate since the situation in which the
        ! PSF partially overlaps some convolver cells is not detected. But
        ! it will have to do for now.

        call CAL_releaseMemory(psfImage)
      end do

      i = ((radialBin - 1) * numAxialBins) + 1 ! ie, same as for 1st axialBin
      ! this radialBin. This splices up the indexImage over the phi=0/phi=2pi
      ! join.

      phiLo = phiBinLoEdges(numAxialBins+1)
      where(thetaMap >= thetaArcsecLo .and. phiMap >= phiLo) indexImage = i
    end do

    call CAL_setState(accuracyLevel=oldAccuracyLevel)

  end subroutine makeConvolvers

  !---------------------------------------------------------------------
  subroutine getPsfImagePixelCorners_old(psfThetaArcsec, psfPhi, psfPixelSizeMm&
  , psfImage, wcs, wcsType&
  , instrumentId, timeStamp, scAttitude&
  , psfImagePixelCorners, psfCentrePixels)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up for the correct instrument before this
  ! is called.


!**** obsolete. Should be replaced by
!****   call getPsfImagePixelCorners
!****   call psfToDet
!****   call instToTan
!****   call tanToPixels


    implicit none

    real(double),                 intent(in)  :: psfThetaArcsec, psfPhi
    type(PsfBinSizeT),            intent(in)  :: psfPixelSizeMm
    real(single),                 intent(in)  :: psfImage(:,:)
    type(WcsT),                   intent(in)  :: wcs
    character(*),                 intent(in)  :: wcsType
    integer(int32),               intent(in)  :: instrumentId
    real(double),                 intent(in)  :: timeStamp
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    type(Point2dT),               intent(out) :: psfImagePixelCorners(&
                                                   size(psfImage,1)+1,&
                                                   size(psfImage,2)+1)
    type(Point2dT),               intent(out) :: psfCentrePixels
    ! type defined in geometric_types.f90.

    integer(int16) :: xSize, ySize
    real(double) ::&
      thetaVec(1), phiVec(1), centreTanX, centreTanY
!    real(single) ::&
real(double) ::&
      centreDetX, centreDetY
    real(double), pointer ::&
      psfImageCentreXmm(:), psfImageCentreYmm(:), dummyZmm(:)
!    integer(int16), pointer :: xIndices(:), yIndices(:)
integer(int16) :: xIndices(size(psfImage,1)+1), yIndices(size(psfImage,2)+1), i
!    real(single), allocatable ::&
real(double), allocatable ::&
      detXVec(:), detYVec(:)
    real(double), allocatable ::&
      tanXArray(:,:), tanYArray(:,:)
!real(double) ::&
!  tanXArray(size(psfImage,1)+1,size(psfImage,2)+1),&
!  tanYArray(size(psfImage,1)+1,size(psfImage,2)+1)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(psfImage, 1)
    ySize = size(psfImage, 2)
!    allocate(psfImagePixelCorners(xSize+1,ySize+1))

    ! Get the centre of the PSF array in CAMCOORD2:

    thetaVec(1) = psfThetaArcsec
    phiVec(1)   = psfPhi
    call CAL_telCoordToCamCoord2(thetaVec, phiVec&
      , psfImageCentreXmm, psfImageCentreYmm, dummyZmm)

    centreDetX = psfImageCentreXmm(0) * mmToDetUnits
    centreDetY = psfImageCentreYmm(0) * mmToDetUnits

    call CAL_releaseMemory(psfImageCentreXmm, psfImageCentreYmm, dummyZmm)

    ! Now do the whole array. However there is no need to deal in arrays
    ! until the final conversion from det coord to sky tangent coords,
    ! because the psf image is aligned with the detx/y coordinate axes.
    ! Hence just work with x-axis and y-axis vectors until that point.

!    xIndices => indicize(xSize+1_int16)
!    yIndices => indicize(ySize+1_int16)
xIndices = (/(i, i = 1, xSize+1)/)
yIndices = (/(i, i = 1, ySize+1)/)

    select case(instrumentId)
    case(EMOS1)
      allocate(detXVec(ySize+1), detYVec(xSize+1))
      detXVec = (ySize / 2.0 - yIndices) * psfPixelSizeMm%binY
      detYVec = (xSize / 2.0 - xIndices) * psfPixelSizeMm%binX

    case(EMOS2)
      allocate(detXVec(xSize+1), detYVec(ySize+1))
      detXVec = (xIndices - xSize / 2.0) * psfPixelSizeMm%binX
      detYVec = (ySize / 2.0 - yIndices) * psfPixelSizeMm%binY

    case(EPN)
      allocate(detXVec(xSize+1), detYVec(ySize+1))
      detXVec = (xSize / 2.0 - xIndices) * psfPixelSizeMm%binX
      detYVec = (yIndices - ySize / 2.0) * psfPixelSizeMm%binY

    case default
call error('badInstrument',"Only EPIC instruments allowed.")
    end select

!    deallocate(xIndices, yIndices)

    detXVec = detXVec * mmToDetUnits + centreDetX
    detYVec = detYVec * mmToDetUnits + centreDetY

    ! Now we have to go from vectors to arrays.

allocate(tanXArray(xSize+1,ySize+1), tanYArray(xSize+1,ySize+1))
    if (wcsType .eq. 'XY') then
!      allocate(tanXArray(xSize+1,ySize+1), tanYArray(xSize+1,ySize+1))

      if (instrumentId .eq. EMOS1) then
!        call instToTan(tanXArray, tanYArray, wcs%x%refWorld&
!          , wcs%y%refWorld, scAttitude&
!          , timeStamp, spread(detXVec, 1, xSize+1)&
!          , spread(detYVec, 2, ySize+1))
call detToTan(spread(detXVec, 1, xSize+1)&
  , spread(detYVec, 2, ySize+1), wcs%x%refWorld*PiOn180&
  , wcs%y%refWorld*PiOn180, scAttitude&
  , timeStamp, tanXArray, tanYArray)

      else
!        call instToTan(tanXArray, tanYArray, wcs%x%refWorld&
!          , wcs%y%refWorld, scAttitude&
!          , timeStamp, spread(detXVec, 2, ySize+1)&
!          , spread(detYVec, 1, xSize+1))
call detToTan(spread(detXVec, 2, ySize+1)&
  , spread(detYVec, 1, xSize+1), wcs%x%refWorld*PiOn180&
  , wcs%y%refWorld*PiOn180, scAttitude&
  , timeStamp, tanXArray, tanYArray)
      end if

      ! Finally, convert from sky tangents to sky pixel coords:

!      psfImagePixelCorners%x = tanToPixelsAxis(wcs%x, tanXArray) 
!      psfImagePixelCorners%y = tanToPixelsAxis(wcs%y, tanYArray) 
call tanToPixels(wcs, tanXArray, tanYArray&
  , psfImagePixelCorners%x, psfImagePixelCorners%y) ! in ssclib/wcs_aux

!      call instToTan(centreTanX, centreTanY, wcs%x%refWorld&
!        , wcs%y%refWorld, scAttitude&
!        , timeStamp, centreDetX, centreDetY)
call detToTan(centreDetX, centreDetY, wcs%x%refWorld*PiOn180&
 , wcs%y%refWorld*PiOn180, scAttitude&
 , timeStamp, centreTanX, centreTanY)

!      psfCentrePixels%x = tanToPixelsAxis(wcs%x, centreTanX) 
!      psfCentrePixels%y = tanToPixelsAxis(wcs%y, centreTanY) 
call tanToPixels(wcs, centreTanX, centreTanY&
  , psfCentrePixels%x, psfCentrePixels%y) ! in ssclib/wcs_aux

!deallocate(tanXArray, tanYArray)

    else if (wcsType .eq. 'DETXY') then
!pixelSizeXWorld = sqrt(wcs%x%matrixOnAxis * wcs%x%matrixOnAxis&
! + wcs%x%matrixOffAxis * wcs%x%matrixOffAxis)
!pixelSizeYWorld = sqrt(wcs%y%matrixOnAxis * wcs%y%matrixOnAxis&
! + wcs%y%matrixOffAxis * wcs%y%matrixOffAxis)
!!*** this is only valid if there is no skew. But such should be the case for this pseudo-wcs defining the detxy-to-image-pixel relationship... I hope..?

      if (instrumentId .eq. EMOS1) then
!        psfImagePixelCorners%x = (spread(detXVec, 1, xSize+1)-wcs%x%refWorld)&
!          / wcs%x%pixelSizeWorld + wcs%x%refPixels
!        psfImagePixelCorners%y = (spread(detYVec, 2, ySize+1)-wcs%y%refWorld)&
!          / wcs%y%pixelSizeWorld + wcs%y%refPixels
!!*** Note no 'abs()'

tanXArray = (spread(detXVec, 1, xSize+1) - wcs%x%refWorld) * detUnitArcsec / 3600.0 * PiOn180
tanYArray = (spread(detYVec, 2, ySize+1) - wcs%y%refWorld) * detUnitArcsec / 3600.0 * PiOn180

call tanToPixels(wcs, tanXArray, tanYArray&
  , psfImagePixelCorners%x, psfImagePixelCorners%y) ! in ssclib/wcs_aux

      else
!        psfImagePixelCorners%x = (spread(detXVec, 2, ySize+1)-wcs%x%refWorld)&
!          / wcs%x%pixelSizeWorld + wcs%x%refPixels
!        psfImagePixelCorners%y = (spread(detYVec, 1, xSize+1)-wcs%y%refWorld)&
!          / wcs%y%pixelSizeWorld + wcs%y%refPixels
!!*** Note no 'abs()'

tanXArray = (spread(detXVec, 2, ySize+1) - wcs%x%refWorld) * detUnitArcsec / 3600.0 * PiOn180
tanYArray = (spread(detYVec, 1, xSize+1) - wcs%y%refWorld) * detUnitArcsec / 3600.0 * PiOn180

call tanToPixels(wcs, tanXArray, tanYArray&
  , psfImagePixelCorners%x, psfImagePixelCorners%y) ! in ssclib/wcs_aux

      end if

!      psfCentrePixels%x = (centreDetX - wcs%x%refWorld)&
!        / wcs%x%pixelSizeWorld + wcs%x%refPixels
!      psfCentrePixels%y = (centreDetY - wcs%y%refWorld)&
!        / wcs%y%pixelSizeWorld + wcs%y%refPixels
!!*** Note no 'abs()'

centreTanX = (centreDetX - wcs%x%refWorld) * detUnitArcsec / 3600.0 * PiOn180
centreTanY = (centreDetY - wcs%y%refWorld) * detUnitArcsec / 3600.0 * PiOn180

call tanToPixels(wcs, centreTanX, centreTanY&
  , psfCentrePixels%x, psfCentrePixels%y) ! in ssclib/wcs_aux

    else
call error('bug0',"Please contact the code developer.")
! because this should already have been screened out.
    end if

!**** can rearrange it so that tanX/YArray and centreTanX/Y are calculated within the select but psfImagePixelCorners and psfCentrePixels outside it.

deallocate(tanXArray, tanYArray)
deallocate(detXVec, detYVec)

  end subroutine getPsfImagePixelCorners_old

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module econvolverprep_mod
